"""Stabilizer verification tools for quantum error correction.

This module provides utilities for verifying stabilizer codes and analyzing
their properties, including stabilizer group verification, code distance
calculation, and logical operator validation.
"""

# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from itertools import combinations, product
from typing import TYPE_CHECKING

import pecos as pc
from pecos.circuits import QuantumCircuit

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

    from pecos.protocols import SimulatorProtocol
    from pecos.typing import LogicalOpInfo, StabilizerVerificationResult

# TODO: NEED TO ADD SIGN TRACKING TO DESTABILIZERS TO GET THE RIGHT SIGN FOR LOGICAL Xs


class VerifyStabilizers:
    """Used to define a stabilizer QECC."""

    def __init__(self) -> None:
        """Initialize the VerifyStabilizers instance.

        Sets up the circuit simulator and initializes empty data structures
        for stabilizer checks, logical operators, and qubit tracking.
        """
        self.circ_sim = pc.simulators.SparseSimPy

        self.checks = []
        self.logical_zs = []
        self.logical_xs = []
        self.logical_zs_defined = []  # User chosen logical Zs TODO: ...
        self.logical_xs_defined = []  # User chosen logical Xs TODO: ...
        self.logical_zs_reference = {}
        self.logical_xs_reference = {}

        self.data_qubits = set()
        self.ancilla_qubits = set()
        self.circuit = None

        self.check_row_x = None
        self.check_row_z = None
        self.check_col_x = None
        self.check_col_z = None

        # stabilizer ids:
        self.data_gens = None
        self.ancilla_gens = None
        self.logical_gens = None

        # distance
        self.dist = None

        self.state = None

    def check(self, paulis: str | Sequence[str], qubits: Sequence[int]) -> None:
        """Check if the given Pauli operators stabilize the state.

        Args:
            paulis: Sequence of Pauli operators to check (e.g., ['X', 'Z', 'Y']).
            qubits: Sequence of qubit indices corresponding to the Pauli operators.

        Returns: None

        """
        if not qubits:
            msg = "No qubit ids given."
            raise Exception(msg)

        check_string = "check(" + str(paulis) + ", " + str(qubits) + ")"

        if isinstance(paulis, str):
            if paulis not in {"X", "x", "Y", "y", "Z", "z"}:
                msg = 'Paulis should be "X", "Y" or "Z"!'
                raise Exception(msg)

            paulis_new = [paulis for _ in qubits]
            paulis = paulis_new

        if not isinstance(paulis, str) and len(paulis) != len(qubits):
            msg = "Number of Paulis and qubits do not match!!!"
            raise Exception(msg)

        self.checks.append((paulis, qubits, check_string))
        self.data_qubits.update(qubits)

    def logicalz(self, paulis: str | Sequence[str], qubits: Sequence[int]) -> None:
        """Used to define logical Z.

        Args:
        ----
            paulis: Either a single Pauli string ('X', 'Y', or 'Z') or a list of Pauli operators.
            qubits: List of qubit indices where the logical Z operator acts.
        """
        if not qubits:
            msg = "No qubit ids given."
            raise Exception(msg)

        logical_string = "check(" + str(paulis) + ", " + str(qubits) + ")"

        if isinstance(paulis, str):
            if paulis not in {"X", "x", "Y", "y", "Z", "z"}:
                msg = 'Paulis should be "X", "Y" or "Z"!'
                raise Exception(msg)

            paulis_new = [paulis for _ in qubits]
            paulis = paulis_new

        if not isinstance(paulis, str) and len(paulis) != len(qubits):
            msg = "Number of Paulis and qubits do not match!!!"
            raise Exception(msg)

        self.logical_zs.append((paulis, qubits, logical_string))

    def logicalx(self, paulis: str | Sequence[str], qubits: Sequence[int]) -> None:
        """Used to define logical X.

        Args:
        ----
            paulis: Either a single Pauli string ('X', 'Y', or 'Z') or a list of Pauli operators.
            qubits: List of qubit indices where the logical X operator acts.
        """
        if not qubits:
            msg = "No qubit ids given."
            raise Exception(msg)

        logical_string = "check(" + str(paulis) + ", " + str(qubits) + ")"

        if isinstance(paulis, str):
            if paulis not in {"X", "x", "Y", "y", "Z", "z"}:
                msg = 'Paulis should be "X", "Y" or "Z"!'
                raise Exception(msg)

            paulis_new = [paulis for _ in qubits]
            paulis = paulis_new

        if not isinstance(paulis, str) and len(paulis) != len(qubits):
            msg = "Number of Paulis and qubits do not match!!!"
            raise Exception(msg)

        self.logical_xs.append((paulis, qubits, logical_string))

    def num_logical_qubits(self) -> int:
        """Calculate the number of logical qubits in the stabilizer code.

        Returns:
            Number of logical qubits (data qubits minus stabilizer checks).
        """
        return len(self.data_qubits) - len(self.checks)

    def generators(
        self,
        *,
        print_y: bool = True,
        verbose: bool = True,
    ) -> tuple[
        list[dict[str, set[int]]],
        list[dict[str, set[int]]],
        list[str],
        list[str],
    ]:
        """Evaluates the stabilizer generators that have been supplied via the `check` method.

        Args:
        ----
            print_y: If True, includes Y operators in the output (otherwise converts to X and Z).
            verbose: If True, prints detailed information about the generators.
        """
        if self.circuit is None:
            msg = "Must compile circuits first!"
            raise Exception(msg)

        state = self.state
        z, x, stab_strings, destab_strings = self.get_info(
            state,
            print_y=print_y,
            verbose=verbose,
        )

        return z, x, stab_strings, destab_strings

    def _check_all_labels(self) -> None:
        """This checks to see that all the consecutive qubit ids have been used and none are missing."""
        qubit_labels = set()

        checks = self.checks
        for check in checks:
            _, qs, _ = check
            qubit_labels.update(qs)

        largest_labels = max(qubit_labels)
        labels_should_have = set(range(largest_labels + 1))

        dont_have = labels_should_have - qubit_labels

        if dont_have:
            msg = f"Qubit ids missing: {dont_have}"
            raise Exception(msg)

    def _check2rowcol(self) -> None:
        """Creates row and column matrices."""
        checks = self.checks

        num_checks = len(checks)
        row_x = [set() for _ in range(num_checks)]
        row_z = [set() for _ in range(num_checks)]
        col_x = [set() for _ in range(self.num_data_qubits)]
        col_z = [set() for _ in range(self.num_data_qubits)]

        for stab_id, check in enumerate(checks):
            ps, qs, _ = check

            for p, q in zip(ps, qs, strict=False):
                if p in {"X", "x"}:
                    row_x[stab_id].add(q)
                    col_x[q].add(stab_id)

                elif p in {"Z", "z"}:
                    row_z[stab_id].add(q)
                    col_z[q].add(stab_id)

                elif p in {"Y", "y"}:
                    row_x[stab_id].add(q)
                    row_z[stab_id].add(q)
                    col_x[q].add(stab_id)
                    col_z[q].add(stab_id)

        self.check_row_x = row_x
        self.check_row_z = row_z
        self.check_col_x = col_x
        self.check_col_z = col_z

    def _check_commute(self) -> bool:
        """Checks to see that all the stabilizer generators commute.

        Returns:
            Returns bool value if all the checks commute or not.
        """
        row_x = self.check_row_x
        row_z = self.check_row_z
        col_x = self.check_col_x
        col_z = self.check_col_z

        for stab_id in range(len(self.checks)):
            anti_zs = set()
            for q in row_x[stab_id]:
                anti_zs ^= col_z[q]

            anti_xs = set()
            for q in row_z[stab_id]:
                anti_xs ^= col_x[q]

            anti = anti_xs ^ anti_zs
            anti.discard(stab_id)

            if anti:
                print("\nChecks anticommute!")
                print("\nCheck:")
                for s in anti:
                    print(self.checks[s][2])
                print("\nanticommutes with:")
                print(self.checks[stab_id][2])

                msg = "Checks anticommute!"
                raise Exception(msg)
        return True

    def compile(self) -> None:
        """Checks commutation relations and creates a circuit to measure the checks."""
        if self.circuit:
            msg = "Measurement encoding-circuit has already been compiled!"
            raise Exception(msg)

        # Check the qubit ids
        self._check_all_labels()

        # Create row and column matrices.
        self._check2rowcol()

        # Checks that all the stabilizer generators (checks) commute.
        self._check_commute()

        # Create check circuits:
        # ----------------------
        ancilla_qubits = set()
        qc = QuantumCircuit()

        ancilla_id = sorted(self.data_qubits)[-1]

        for ps, qs, _ in self.checks:
            ancilla_id += 1
            ancilla_qubits.add(ancilla_id)
            qc.append("init |+>", {ancilla_id})

            for p, q in zip(ps, qs, strict=False):
                symbol = None

                if p in {"X", "x"}:
                    symbol = "CNOT"
                elif p in {"Z", "z"}:
                    symbol = "CZ"
                elif p in {"Y", "y"}:
                    symbol = "CY"

                qc.append(symbol, {(ancilla_id, q)})

            qc.append("measure X", {ancilla_id}, random_outcome=0)

        self.ancilla_qubits = ancilla_qubits

        self.circuit = qc

        # Run circuits
        # ------------
        # Separate the checks, logical stabilizers, and ancilla stabilizers.
        circuit = self.circuit
        state = pc.simulators.SparseSimPy(self.num_qubits)
        state.run_circuit(circuit)
        self.get_info(state, verbose=False)
        self.state = state

        self._verify_checks()
        self._check_logical_commute()

    def _check_logical_commute(self) -> None:
        logical_z_col_x = [set() for _ in range(self.num_data_qubits)]
        logical_z_col_z = [set() for _ in range(self.num_data_qubits)]
        logical_z_row_x = [set() for _ in range(len(self.logical_zs))]
        logical_z_row_z = [set() for _ in range(len(self.logical_zs))]

        logical_x_col_x = [set() for _ in range(self.num_data_qubits)]
        logical_x_col_z = [set() for _ in range(self.num_data_qubits)]
        logical_x_row_x = [set() for _ in range(len(self.logical_xs))]
        logical_x_row_z = [set() for _ in range(len(self.logical_xs))]

        for i, (ps, qs, _) in enumerate(self.logical_zs):
            for p, q in zip(ps, qs, strict=False):
                if p in {"X", "Y"}:
                    logical_z_col_x[q].add(i)
                    logical_z_row_x[i].add(q)

                if p in {"Z", "Y"}:
                    logical_z_col_z[q].add(i)
                    logical_z_row_z[i].add(q)

        for i, (ps, qs, _) in enumerate(self.logical_xs):
            for p, q in zip(ps, qs, strict=False):
                if p in {"X", "Y"}:
                    logical_x_col_x[q].add(i)
                    logical_x_row_x[i].add(q)

                if p in {"Z", "Y"}:
                    logical_x_col_z[q].add(i)
                    logical_x_row_z[i].add(q)

        for s in range(len(self.logical_zs)):
            anti_zs = set()
            for q in logical_z_row_x[s]:
                anti_zs ^= logical_z_col_z[q]

            anti_xs = set()
            for q in logical_z_row_z[s]:
                anti_xs ^= logical_z_col_x[q]

            anti = anti_xs ^ anti_zs
            anti.discard(s)

            if anti:
                print("\nLogical Zs anticommute!")
                print("\nLogical Zs:")
                for i in anti:
                    print(self.logical_zs[i][2])
                print("\nanticommutes with:")
                print(self.logical_zs[s][2])

                msg = "Logical Zs anticommute!"
                raise Exception(msg)

        for s in range(len(self.logical_xs)):
            anti_zs = set()
            for q in logical_x_row_x[s]:
                anti_zs ^= logical_x_col_z[q]

            anti_xs = set()
            for q in logical_x_row_z[s]:
                anti_xs ^= logical_x_col_x[q]

            anti = anti_xs ^ anti_zs
            anti.discard(s)

            if anti:
                print("\nLogical Xs anticommute!")
                print("\nLogical Xs:")
                for i in anti:
                    print(self.logical_xs[i][2])
                print("\nanticommutes with:")
                print(self.logical_xs[s][2])

                msg = "Logical Xs anticommute!"
                raise Exception(msg)

        # So far checked that all the logical Zs and logical Xs commute with themselves...
        # - Next check that they commute with the stabilizers...
        # - Then find if the there are anti-commuting pairs of logical Zs and Xs
        # - Then search for the logical operators and refactor...
        # This step might require switching logical Xs and Zs... Might be a bit complicated... as we can modify
        # "logical Xs" with destabilizers and swap those... So need to do a search through all stabilizers and
        # destabilizers and then determine if the required multiplication is valid with fix stabiliziers and whatever
        # has been fixed for the logical operators...

    def _verify_checks(self) -> bool:
        # Stabilizers:
        checks = []

        for strings, qids, _ in self.checks:
            check_dict = {}
            for pauli, q in zip(strings, qids, strict=False):
                if pauli == "X":
                    qset = check_dict.setdefault("X", set())
                elif pauli == "Z":
                    qset = check_dict.setdefault("Z", set())
                else:
                    qset = check_dict.setdefault("Y", set())
                qset.add(q)
            checks.append(check_dict)

        checks2 = []
        for i in range(len(self.checks)):
            xs = self.check_row_x[i]
            zs = self.check_row_z[i]

            stab_dict = {}

            if xs - zs:
                stab_dict["X"] = xs - zs

            if zs - xs:
                stab_dict["Z"] = zs - xs

            if xs & zs:
                stab_dict["Y"] = xs & zs

            checks2.append(stab_dict)

        if checks != checks2:
            print(
                "WARNING: PECOS didn't refactor the stabilizers into the checks supplied!",
            )

        return checks != checks2

    def eval(self, *, verbose: bool = False) -> StabilizerVerificationResult:
        """Evaluate the stabilizer code verification.

        Args:
            verbose: Whether to print detailed output during evaluation.

        Returns:
            Verification result containing success status and error details.
        """
        if self.circuit is None:
            self.compile()

        z, x, _, destab_strings = self.generators(verbose=verbose)

        if self.dist is None:
            self.distance(verbose=verbose)

        # Stabilizers:
        checks = []

        for strings, qids, _ in self.checks:
            check_dict = {}
            for pauli, q in zip(strings, qids, strict=False):
                if pauli == "X":
                    qset = check_dict.setdefault("X", set())
                elif pauli == "Z":
                    qset = check_dict.setdefault("Z", set())
                else:
                    qset = check_dict.setdefault("Y", set())
                qset.add(q)
            checks.append(check_dict)

        # Destabilizers:
        destabs = []
        for i in self.data_gens:
            destab_dict = {}
            for j in range(len(destab_strings[i]) - self.num_ancilla_qubits):
                pauli = destab_strings[i][j]
                if pauli == "X":
                    qset = destab_dict.setdefault("X", set())
                elif pauli == "Z":
                    qset = destab_dict.setdefault("Z", set())
                elif pauli == "Y":
                    qset = destab_dict.setdefault("Y", set())
                else:
                    continue
                qset.add(j - 2)
            destabs.append(destab_dict)

        output_dict = {
            "num_datas": self.num_data_qubits,
            "num_logical_qubits": self.num_logical_qubits(),
            "distance": self.dist,
            "[[n, k, d]]": f"[[{self.num_data_qubits}, {self.num_logical_qubits()}, {self.dist}]]",
            "checks": checks,
            "destabilizers": destabs,
            "logical_xs": x,
            "logical_zs": z,
        }

        self.logical_xs_reference = {}
        self.logical_zs_reference = {}

        for i, xi in enumerate(x):
            self.logical_xs_reference["X" + str(i)] = xi

        for i, zi in enumerate(z):
            self.logical_zs_reference["Z" + str(i)] = zi

        return output_dict

    @property
    def num_data_qubits(self) -> int:
        """Get the number of data qubits.

        Returns:
            Number of data qubits in the stabilizer code.
        """
        return len(self.data_qubits)

    @property
    def num_ancilla_qubits(self) -> int:
        """Get the number of ancilla qubits.

        Returns:
            Number of ancilla qubits in the stabilizer code.
        """
        return len(self.ancilla_qubits)

    @property
    def num_qubits(self) -> int:
        """Get the total number of qubits.

        Returns:
            Total number of qubits (data + ancilla).
        """
        return len(self.data_qubits) + len(self.ancilla_qubits)

    def refactor(self, state: SimulatorProtocol) -> None:
        """Refactor the stabilizer state to match the expected generators.

        Args:
            state: Simulator state to refactor.
        """
        found_stab_ids = set()

        refactor_things = list(self.checks)
        refactor_things.extend(self.logical_zs)
        # TODO: NEED TO REFACTOR THE DESTABILIZER OF LOGICAL Z TO GET THE RIGHT LOGICAL X.....

        for ps, qs, _ in refactor_things:
            xs = set()
            zs = set()

            for p, q in zip(ps, qs, strict=False):
                if p in {"X", "x"}:
                    xs.add(q)
                elif p in {"Z", "z"}:
                    zs.add(q)
                elif p in {"Y", "y"}:
                    xs.add(q)
                    zs.add(q)

            try:
                found, stab_id = state.refactor(
                    xs,
                    zs,
                    choose=0,
                    protected=found_stab_ids,
                )
            except IndexError:
                xonly = xs - zs
                zonly = zs - xs
                ys = xs & zs
                msg = f"IndexError.\nThe stabilizer {{'X': {xonly}, 'Y': {ys}, 'Z': {zonly}}} is likely redundant!"
                raise Exception(msg) from IndexError

            found_stab_ids.add(stab_id)

            if not found:
                msg = "Could not find check:"
                raise Exception(msg, (ps, qs))

        for q in self.ancilla_qubits:
            found, stab_id = state.refactor(
                {q},
                set(),
                choose=-1,
                protected=found_stab_ids,
            )
            found_stab_ids.add(stab_id)

            if not found:
                msg = f"Could not find ancilla {q}"
                raise Exception(msg)

    def get_check_ancilla(
        self,
    ) -> tuple[list[tuple[set[int], set[int]]], list[tuple[set[int], set[int]]]]:
        """Get check and ancilla operator sets.

        Returns:
            Tuple containing lists of (X set, Z set) tuples for checks and ancillas.
        """
        check_tuples = []
        ancilla_tuples = []

        for ps, qs, _ in self.checks:
            xs = set()
            zs = set()

            for p, q in zip(ps, qs, strict=False):
                if p in {"X", "x"}:
                    xs.add(q)
                elif p in {"Z", "z"}:
                    zs.add(q)
                elif p in {"Y", "y"}:
                    xs.add(q)
                    zs.add(q)

            check_tuples.append((xs, zs))

        ancilla_tuples.extend(({q}, set()) for q in self.ancilla_qubits)

        return check_tuples, ancilla_tuples

    def get_info(
        self,
        state: SimulatorProtocol,
        stop_search: int = 1000,
        *,
        verbose: bool = True,
        print_y: bool = False,
    ) -> tuple[
        list[dict[str, set[int]]],
        list[dict[str, set[int]]],
        list[str],
        list[str],
    ]:
        """Get stabilizer information from the quantum state.

        Args:
            state: Simulator state to analyze.
            stop_search: Maximum number of refactoring attempts.
            verbose: Whether to print detailed information.
            print_y: Whether to include Y operators in output.

        Returns:
            Tuple of logical Z operators, logical X operators, stabilizer strings, destabilizer strings.
        """
        if self.circuit is None:
            return Exception("Must run `compile()` first!")

        self.refactor(state)
        stab_strs, destab_strs = state.print_stabs(
            verbose=False,
            print_y=print_y,
            print_destabs=True,
        )

        num_ancillas = len(self.ancilla_qubits)

        num_logical = self.num_logical_qubits()
        num_checks = len(self.checks)

        if verbose:
            print(f"Number of data qubits: {self.num_data_qubits}")
            print(f"Number of checks: {num_checks}")
            print(f"Number of logical qubits: {num_logical}")

        check_tuples, ancilla_tuples = self.get_check_ancilla()
        # determine the gen_id of the checks and logicals
        check_gens = []
        logical_gens = []
        ancilla_gens = []

        missing_checks = list(check_tuples)
        missing_ancillas = list(ancilla_tuples)
        notmatched_gens = list(range(state.num_qubits))

        found_all = False
        search_count = 0
        while not found_all:
            if verbose:
                print("----")
            for g, gtuple in enumerate(
                zip(state.stabs.row_x, state.stabs.row_z, strict=False),
            ):
                if gtuple in missing_checks:
                    missing_checks.remove(gtuple)
                    try:
                        notmatched_gens.remove(g)
                    except ValueError:
                        msg = f"list.remove(x): x not in list.\nThe stabilizer {gtuple!s} is likely redundant!"
                        raise Exception(
                            msg,
                        ) from ValueError

                    check_gens.append(g)
                elif gtuple in missing_ancillas:
                    missing_ancillas.remove(gtuple)
                    notmatched_gens.remove(g)
                    ancilla_gens.append(g)

            if len(notmatched_gens) == num_logical:
                logical_gens = notmatched_gens
                found_all = True
            else:
                for xs, zs in missing_checks:
                    state.refactor(xs, zs, choose=0, prefer=notmatched_gens)
                state.print_stabs(verbose=False, print_y=print_y, print_destabs=True)

            if search_count == stop_search:
                msg = "Can not refactor properly!"
                raise Exception(msg)
            search_count += 1

        self.data_gens = set(check_gens)
        self.ancilla_gens = set(ancilla_gens)
        self.logical_gens = set(logical_gens)

        if verbose:
            if len(check_gens) != num_checks:
                print("Found:", check_gens)
                print("Want:", check_tuples)
                msg = f"Did not find the correct number of stabilizer generators. {len(check_gens)}/{num_checks}"
                raise Exception(msg)

            if len(logical_gens) != num_logical:
                print("Found:", logical_gens)
                msg = f"Did not find the correct number of logical generators. {len(logical_gens)}/{num_logical}"
                raise Exception(msg)

            print("\nStabilizer generators:")
            for gen in check_gens:
                print(stab_strs[gen][: len(stab_strs[gen]) - num_ancillas])

            print("\nDestabilizer generators:")
            for gen in check_gens:
                print(destab_strs[gen][: len(destab_strs[gen]) - num_ancillas])

            print("\nLogical operators:")

            for i, gen in enumerate(logical_gens):
                print(f"\n. Logical Z #{i + 1!s}:")
                print(stab_strs[gen][: len(stab_strs[gen]) - num_ancillas])
                print(f". Logical X #{i + 1!s}:")
                print(destab_strs[gen][: len(destab_strs[gen]) - num_ancillas])

        logical_z_strings = []
        logical_x_strings = []

        for gen in logical_gens:
            z_string = stab_strs[gen][: len(stab_strs[gen]) - num_ancillas]
            x_string = destab_strs[gen][: len(destab_strs[gen]) - num_ancillas]

            check_dict = {}
            for q, pauli in enumerate(z_string):
                if pauli == "X":
                    qset = check_dict.setdefault("X", set())
                elif pauli == "Z":
                    qset = check_dict.setdefault("Z", set())
                elif pauli == "Y":
                    qset = check_dict.setdefault("Y", set())
                else:
                    continue
                qset.add(q - 2)
            logical_z_strings.append(check_dict)

            check_dict = {}
            for q, pauli in enumerate(x_string):
                if pauli == "X":
                    qset = check_dict.setdefault("X", set())
                elif pauli == "Z":
                    qset = check_dict.setdefault("Z", set())
                elif pauli == "Y":
                    qset = check_dict.setdefault("Y", set())
                else:
                    continue

                qset.add(q - 2)
            logical_x_strings.append(check_dict)

        return logical_z_strings, logical_x_strings, stab_strs, destab_strs

    def distance(
        self,
        *,
        css: bool = False,
        verbose: bool = True,
    ) -> tuple[set[int], set[int]] | None:
        """Checks the distance of the code."""
        if self.circuit is None:
            msg = "Must compile circuits first!"
            raise Exception(msg)

        qudit_set = self.data_qubits

        state = self.state
        found = self._dist_mode_smallest(state, qudit_set, css=css, verbose=verbose)

        if verbose and found:
            xs, zs = found
            distance = len(xs | zs)

            print(
                f"\nThis is a [[{self.num_data_qubits}, {self.num_logical_qubits()}, {distance}]] code.",
            )

        if not found:
            print(
                "No logical errors found... Checks might describe a stabilizer state.",
            )
            return None
        xs, zs = found
        self.dist = len(xs | zs)
        return found

    def _dist_mode_smallest(
        self,
        state: SimulatorProtocol,
        qudit_set: set[int],
        *,
        css: bool = False,
        verbose: bool = True,
        start_len: int | None = None,
        end_len: int | None = None,
        list_ops: bool = False,
    ) -> Generator[tuple[set, set], None, None]:
        """Determine if a logical error can be found by starting with the smallest weight errors.

        Args:
        ----
            state: The quantum state to check for logical errors.
            qudit_set: Set of qudit indices to consider for errors.
            css: If True, only checks CSS (Calderbank-Shor-Steane) type errors.
            verbose: If True, prints progress and found logical operators.
            start_len: Starting weight of errors to check (default: 1).
            end_len: Maximum weight of errors to check (default: number of qudits).
            list_ops: If True, returns a list of all logical operators found.

        """
        ops = []

        if start_len is None:
            start_len = 1

        if end_len is None:
            end_len = len(qudit_set)

        for lenq in range(start_len, end_len + 1):
            if verbose:
                print(f"Checking Paulis of weight {lenq}...")

            for xs, zs in self.gen_errors(qudit_set, lenq, lenq, css=css):
                if self._is_logical_error(state, xs, zs):
                    if verbose:
                        print(f"Logical operator found: Xs - {xs} Zs - {zs}")

                    if list_ops:
                        ops.append({"X": xs, "Z": zs})
                    else:
                        return xs, zs

        return ops

    def gen_errors(
        self,
        qubits: set[int] | Sequence[int],
        min_errors: int = 1,
        *,
        max_errors: bool | int = False,
        css: bool = False,
    ) -> Generator[tuple[set[int], set[int]], None, None]:
        """Generate error patterns for testing stabilizer codes.

        Args:
        ----
            qubits (set of int): Set of qubit indices to generate errors on.
            min_errors (int): Minimum number of errors to generate.
            max_errors (bool, int): Maximum number of errors to generate. False for no limit.
            css (bool): If True, generate only CSS-compatible errors (X and Z only).

        """
        paulis = ("X", "Z", "Y")

        num_qubits = len(qubits)

        for i in range(min_errors, num_qubits + 1):
            if max_errors and i > max_errors:
                break

            xs = next(product(("X",), repeat=i))
            zs = next(product(("Z",), repeat=i))

            xzs = [xs, zs]

            for b in combinations(qubits, i):
                for ps in xzs:
                    x_set = set()
                    z_set = set()
                    for p, q in zip(ps, b, strict=False):
                        if p == "X":
                            x_set.add(q)
                        else:
                            z_set.add(q)
                    yield x_set, z_set

            if not css:
                for a in product(paulis, repeat=i):
                    if a in {xs, zs}:
                        continue

                    for b in combinations(qubits, i):
                        x_set = set()
                        z_set = set()
                        for p, q in zip(a, b, strict=False):
                            if p == "X":
                                x_set.add(q)
                            elif p == "Z":
                                z_set.add(q)
                            else:
                                x_set.add(q)
                                z_set.add(q)
                        yield x_set, z_set

    def _is_logical_error(
        self,
        state: SimulatorProtocol,
        xs: set[int],
        zs: set[int],
    ) -> bool:
        # A trivial error anticommutes with the checks. (Might or might not anticommute with the logical stabilizers)
        # A logical error commutes with the checks and is not a product of checks.

        # Does the error anticommute with the checks?
        x_anticoms = set()
        z_anticoms = set()
        for q in xs:
            x_anticoms ^= state.stabs.col_z[q]

        for q in zs:
            z_anticoms ^= state.stabs.col_x[q]

        anticoms = x_anticoms ^ z_anticoms
        anticom_logical_zs = self.logical_gens & anticoms
        anticoms -= self.logical_gens

        if anticoms:
            return False
        if anticom_logical_zs:
            # So the error commutes with all the stabilizers
            # Did it anticommute with any logical Z operations? If so... It is a product of logical Xs!
            # (and possibly other things)
            return True
        # Let's see if the error anticommuted with any logical X operators:

        x_anticoms_destabs = set()
        z_anticoms_destabs = set()

        for q in xs:
            x_anticoms_destabs ^= state.destabs.col_z[q]

        for q in zs:
            z_anticoms_destabs ^= state.destabs.col_x[q]

        anticoms_destabs = x_anticoms_destabs ^ z_anticoms_destabs
        anticom_logical_xs = self.logical_gens & anticoms_destabs

        # The error is a product of logical Zs
        return bool(anticom_logical_xs)

    def shortest_logicals(
        self,
        start_weight: int | None = None,
        delta: int = 0,
        *,
        verbose: bool = True,
        css: bool = False,
    ) -> tuple[
        list[LogicalOpInfo],
        dict[str, dict[str, set[int]]],
        dict[str, dict[str, set[int]]],
    ]:
        """Find the shortest logical operators.

        Args:
            start_weight (int): Weight of operators to begin searching.
            delta (int): Method will look for all logical ops with weight =< minimum weight + `delta`.
            verbose (bool): If True, print progress information during the search.
            css (bool): If True, restrict search to CSS-compatible operators (X and Z only).

        Returns:
        -------
            Dictionary of logical ops...

        """
        # if not self.logical_xs_reference and not self.logical_zs_reference:

        if start_weight is None:
            start_weight = self.dist if self.dist is not None else 1

        end_weight = start_weight + delta

        if self.circuit is None:
            msg = "Must compile circuits first!"
            raise Exception(msg)

        qudit_set = self.data_qubits

        end_weight = min(end_weight, len(qudit_set))

        state = self.state
        found = self._dist_mode_smallest(
            state,
            qudit_set,
            css=css,
            verbose=False,
            start_len=start_weight,
            end_len=end_weight,
            list_ops=True,
        )

        xs_labels = sorted(self.logical_xs_reference.keys())
        zs_labels = sorted(self.logical_zs_reference.keys())

        oplist = []

        if found:
            for paulis in found:
                op_product = []
                for xi, op_label in enumerate(xs_labels):
                    if self.op_anticommute(paulis, self.logical_xs_reference[op_label]):
                        op_product.append(zs_labels[xi])

                for zi, op_label in enumerate(zs_labels):
                    if self.op_anticommute(paulis, self.logical_zs_reference[op_label]):
                        op_product.append(xs_labels[zi])

                op_product = sorted(op_product)

                oplist.append(
                    {
                        "X": paulis["X"],
                        "Z": paulis["Z"],
                        "equiv_ops": tuple(op_product),
                    },
                )

        if verbose:
            print("Reference Logical Operators:")
            print("\nLogical Xs:")
            for op_label in xs_labels:
                op = self.logical_xs_reference[op_label]
                print(op_label, op)
            print("\nLogical Zs:")
            for op_label in zs_labels:
                op = self.logical_zs_reference[op_label]
                print(op_label, op)

            print("\nLogical Ops Found:\n")
            for foundop in oplist:
                print(
                    "X - {} Z - {} Equiv Ops - {}".format(
                        foundop["X"],
                        foundop["Z"],
                        foundop["equiv_ops"],
                    ),
                )

        return oplist, self.logical_xs_reference, self.logical_zs_reference

    @staticmethod
    def op_anticommute(op1: dict[str, set[int]], op2: dict[str, set[int]]) -> bool:
        """Check if two Pauli operators anticommute.

        Args:
            op1: First Pauli operator as dictionary with X, Y, Z keys and qubit sets.
            op2: Second Pauli operator as dictionary with X, Y, Z keys and qubit sets.

        Returns:
            True if the operators anticommute, False otherwise.
        """
        return bool(
            (len(op1.get("X", set()) & op2.get("Z", set())) + len(op2.get("X", set()) & op1.get("Z", set()))) % 2,
        )



"""Guppy-only tests that don't require full PECOS installation."""

import pytest
from pecos import get_guppy_backends


def test_guppy_available() -> None:
    """Test if Guppy is available in the environment."""
    try:
        from guppylang import guppy

        @guppy
        def test_func(x: int) -> int:
            return x + 1

        # Function should be a GuppyDefinition
        assert hasattr(test_func, "id") or hasattr(test_func, "compile")

    except ImportError:
        pytest.skip("guppylang not available - install with: uv pip install guppylang")


def test_backend_detection_minimal() -> None:
    """Test backend detection without full PECOS."""
    backends = get_guppy_backends()

    # Should return a dict
    assert isinstance(backends, dict)
    assert "guppy_available" in backends
    assert "rust_backend" in backends



# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""A dummy decoder that gives no recovery (outputs do nothing) given any input."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.circuits import QuantumCircuit
    from pecos.misc.std_output import StdOutput


class DummyDecoder:
    """This decoder is just a simple look-up decoder."""

    def __init__(self) -> None:
        """Initialize the DummyDecoder.

        This decoder provides no recovery operations for any syndrome input.
        """

    @staticmethod
    def decode(
        _measurements: StdOutput,
        **_kwargs: object,
    ) -> list[QuantumCircuit]:
        """Decode measurements and return recovery operations.

        Args:
        ----
            measurements: The stabilizer measurements to decode
            **kwargs: Additional keyword arguments (ignored)

        Returns:
        -------
            Empty list - no recovery operations

        """
        return []



"""Test the sim builder pattern API.

This test demonstrates the sim() API builder pattern for quantum simulations.
"""

from pathlib import Path

import pytest
from guppylang import guppy
from guppylang.std.quantum import cx, h, measure, qubit
from pecos import Guppy, sim
from pecos_rslib import state_vector


def decode_integer_results(results: list[int], n_bits: int) -> list[tuple[bool, ...]]:
    """Decode integer-encoded results back to tuples of booleans."""
    decoded = []
    for val in results:
        bits = [bool(val & (1 << i)) for i in range(n_bits)]
        decoded.append(tuple(bits))
    return decoded


class TestGuppySimBuilder:
    """Test the sim builder pattern."""

    @guppy
    def bell_state() -> tuple[bool, bool]:
        """Create a Bell state."""
        q0, q1 = qubit(), qubit()
        h(q0)
        cx(q0, q1)
        return measure(q0), measure(q1)

    @guppy
    def single_qubit() -> bool:
        """Single qubit in superposition."""
        q = qubit()
        h(q)
        return measure(q)

    def test_basic_build_and_run(self) -> None:
        """Test basic build() and run() pattern."""
        # Build once
        # Run multiple times with same configuration
        results1 = sim(self.bell_state).qubits(10).quantum(state_vector()).run(10)
        results2 = sim(self.bell_state).qubits(10).quantum(state_vector()).run(10)

        # Check format has measurement results
        # Bell state returns tuple, so we should have measurement_0 and measurement_0
        if "measurement_0" in results1 and "measurement_0" in results1:
            # New format with individual measurement keys
            assert len(results1["measurement_0"]) == 10
            assert len(results1["measurement_0"]) == 10
            assert len(results2["measurement_0"]) == 10
            assert len(results2["measurement_0"]) == 10
        else:
            # Fallback to old format
            measurements1 = results1.get("measurements", results1.get("result", []))
            measurements2 = results2.get("measurements", results2.get("result", []))
            assert len(measurements1) == 10
            assert len(measurements2) == 10

    def test_direct_run(self) -> None:
        """Test direct run() without explicit build()."""
        results = sim(self.single_qubit).qubits(10).quantum(state_vector()).run(10).to_dict()

        # Check that we have measurement results
        raw_measurements = results.get("measurements", [])
        measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
        assert len(measurements) == 10
        assert all(r in [0, 1] for r in measurements)

    def test_builder_methods(self) -> None:
        """Test the builder pattern methods of the sim API."""
        builder = (
            sim(self.bell_state)
            .qubits(2)
            .quantum(state_vector())
            .seed(42)
            .workers(2)
            .verbose(True)
            .debug(False)
            .optimize(True)
        )
        sim_obj = builder.build()
        results = sim_obj.run(100)

        measurements = results.get(
            "measurements",
            results.get("measurement_0", results.get("result", [])),
        )
        assert measurements is not None
        assert len(measurements) > 0
        assert len(measurements) == 100  # 100 shots, each with integer-encoded 2 qubits

    def test_seeded_reproducibility(self) -> None:
        """Test that seeded runs are reproducible."""
        # Run with same seed twice
        results1 = sim(self.single_qubit).qubits(10).quantum(state_vector()).seed(12345).run(100)
        results2 = sim(self.single_qubit).qubits(10).quantum(state_vector()).seed(12345).run(100)
        measurements1 = results1.get(
            "measurements",
            results1.get("measurement_0", results1.get("result", [])),
        )
        measurements2 = results2.get(
            "measurements",
            results2.get("measurement_0", results2.get("result", [])),
        )
        assert measurements1 == measurements2

    def test_config_dict(self) -> None:
        """Test configuration via dictionary."""
        # Test seed configuration (most commonly used)
        results = sim(self.bell_state).qubits(10).quantum(state_vector()).seed(42).run(50)
        if "measurement_0" in results:
            assert len(results["measurement_0"]) == 50
            assert len(results["measurement_1"]) == 50
        else:
            measurements = results.get("measurements", results.get("result", []))
            assert len(measurements) == 50

    def test_bell_state_correlation(self) -> None:
        """Test that Bell state results are correlated."""
        results = sim(self.bell_state).qubits(10).quantum(state_vector()).seed(42).run(1000).to_dict()

        # Measurements format is [[m0, m1], [m0, m1], ...]
        raw_measurements = results.get("measurements", [])
        correlated = sum(1 for m in raw_measurements if m[0] == m[1])
        assert correlated == len(
            raw_measurements,
        ), "Bell state should be 100% correlated"

    def test_keep_intermediate_files(self) -> None:
        """Test keeping intermediate compilation files."""
        import shutil

        sim_obj = sim(self.single_qubit).qubits(10).quantum(state_vector()).keep_intermediate_files(True).build()
        assert sim_obj.temp_dir is not None
        assert Path(sim_obj.temp_dir).exists()

        # Check that intermediate files exist
        temp_path = Path(sim_obj.temp_dir)
        ll_files = list(temp_path.glob("*.ll"))
        hugr_files = list(temp_path.glob("*.hugr"))

        assert len(ll_files) > 0, "Should have created LLVM IR file"
        assert len(hugr_files) > 0, "Should have created HUGR file"

        # Run simulation
        results = sim_obj.run(10).to_dict()
        raw_measurements = results.get("measurements", [])
        measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
        assert len(measurements) == 10

        # Files should still exist after run
        assert Path(sim_obj.temp_dir).exists()
        assert ll_files[0].exists()
        assert hugr_files[0].exists()

        # Manually clean up
        shutil.rmtree(sim_obj.temp_dir, ignore_errors=True)


def test_api_demonstration() -> None:
    """Demonstrate the builder pattern API."""
    try:
        from guppylang import guppy
        from guppylang.std.quantum import h, measure, qubit
    except ImportError:
        pytest.skip("Guppy not available")
        return

    @guppy
    def demo_circuit() -> bool:
        """Demo circuit that creates superposition and measures."""
        q = qubit()
        h(q)
        return measure(q)

    # Show builder pattern
    sim_obj = sim(demo_circuit).qubits(10).quantum(state_vector()).seed(42).verbose(True).build()
    results = sim_obj.run(100)
    results.get(
        "measurements",
        results.get("measurement_0", results.get("result", [])),
    )

    # print("\n3. Running 1000 shots with a new builder...")
    # Need to create a new builder since the previous one is consumed
    results = sim(Guppy(demo_circuit)).qubits(10).quantum(state_vector()).seed(42).run(1000)
    results.get(
        "measurements",
        results.get("measurement_0", results.get("result", [])),
    )
    results = sim(Guppy(demo_circuit)).qubits(10).quantum(state_vector()).seed(123).run(50)
    results.get(
        "measurements",
        results.get("measurement_0", results.get("result", [])),
    )


def test_simulation_reset() -> None:
    """Test that reset() returns the simulator and allows re-running."""
    try:
        from guppylang import guppy
        from guppylang.std.quantum import h, measure, qubit
    except ImportError:
        pytest.skip("Guppy not available")
        return

    @guppy
    def superposition() -> bool:
        """Create superposition and measure."""
        q = qubit()
        h(q)
        return measure(q)

    # Build a simulation
    sim_obj = sim(superposition).qubits(10).quantum(state_vector()).seed(42).build()

    # Run once
    results1 = sim_obj.run(100)
    assert len(results1) == 100

    # Reset should return the same simulator object
    returned_sim = sim_obj.reset()
    assert returned_sim is sim_obj

    # Run again after reset - should work and give results
    results2 = sim_obj.run(100)
    assert len(results2) == 100

    # With the same seed, reset should give reproducible results
    sim_obj_seeded = sim(superposition).qubits(10).quantum(state_vector()).seed(42).build()
    results_a = sim_obj_seeded.run(50)
    sim_obj_seeded.reset()
    results_b = sim_obj_seeded.run(50)

    # Both runs should produce the same results since we reset with the same seed
    assert results_a.to_dict() == results_b.to_dict()


def test_reset_returns_to_zero_state() -> None:
    """Test that reset() returns the quantum state to all |0⟩."""
    try:
        from guppylang import guppy
        from guppylang.std.quantum import measure, qubit
    except ImportError:
        pytest.skip("Guppy not available")
        return

    @guppy
    def measure_without_gates() -> bool:
        """Measure a qubit without any gates - should always be 0 in |0⟩ state."""
        q = qubit()
        return measure(q)

    # Build a simulation
    sim_obj = sim(measure_without_gates).qubits(10).quantum(state_vector()).seed(42).build()

    # First run - all measurements should be 0 since qubit starts in |0⟩
    results1 = sim_obj.run(100)
    measurements1 = results1.to_dict().get("measurements", [])
    assert all(
        m == 0 or m == [0] or m == (0,) or m is False for m in measurements1
    ), f"Expected all measurements to be 0, got: {measurements1[:5]}..."

    # Reset the simulator
    sim_obj.reset()

    # After reset, qubit should be back in |0⟩, so all measurements should still be 0
    results2 = sim_obj.run(100)
    measurements2 = results2.to_dict().get("measurements", [])
    assert all(
        m == 0 or m == [0] or m == (0,) or m is False for m in measurements2
    ), f"After reset, expected all measurements to be 0, got: {measurements2[:5]}..."



# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Simple error generator meant to demonstrate a basic error generator that produces errors."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from pecos.circuits.quantum_circuit import QuantumCircuit
from pecos.error_models.class_errors_circuit import ErrorCircuits
from pecos.error_models.parent_class_error_gen import ParentErrorModel

if TYPE_CHECKING:
    from pecos.typing import ErrorParams, GateParams, LocationSet


class GatewiseModel(ParentErrorModel):
    """A simple error generator for the depolarizing model.

    This error generator does not allow much modification of the error model.
    """

    measurements: ClassVar[set[str]] = {"measure X", "measure Y", "measure Z"}
    inits: ClassVar[set[str]] = {
        "init |0>",
        "init |1>",
        "init |+>",
        "init |->",
        "init |+i>",
        "init |-i>",
    }
    two_qubits: ClassVar[set[str]] = {"CNOT", "CZ", "SWAP", "G"}
    one_qubits: ClassVar[set[str]] = {
        "I",
        "X",
        "Y",
        "Z",
        "Q",
        "Qd",
        "R",
        "Rd",
        "S",
        "Sd",
        "H",
        "H1",
        "H2",
        "H3",
        "H4",
        "H5",
        "H6",
        "H+z+x",
        "H-z-x",
        "H+y-z",
        "H-y-z",
        "H-x+y",
        "H-x-y",
        "F1",
        "F1d",
        "F2",
        "F2d",
        "F3",
        "F3d",
        "F4",
        "F4d",
    }

    inits_z: ClassVar[set[str]] = {"init |0>", "init |1>"}
    inits_x: ClassVar[set[str]] = {"init |+>", "init |->"}
    inits_y: ClassVar[set[str]] = {"init |+i>", "init |-i>"}

    def __init__(self) -> None:
        """Initialize a GatewiseModel error generator.

        Sets up gate groups for measurements, preparations, single-qubit gates,
        and two-qubit gates. Inherits generator methods and error function classes
        from the parent generator.
        """
        super().__init__()
        self.gen = self.generator_class()
        self.gen.set_gate_group("measurements", self.measurements)
        self.gen.set_gate_group("preps", self.inits)
        self.gen.set_gate_group("two_qubits", self.two_qubits)
        self.gen.set_gate_group("one_qubits", self.one_qubits)

        # Inherit gen methods
        self.set_gate_error = self.gen.set_gate_error
        self.set_group_error = self.gen.set_group_error
        self.set_default_error = self.gen.set_default_error
        self.set_gate_group = self.gen.set_gate_group

        self.gate_groups = self.gen.gate_groups

        # Inherit error function classes
        self.ErrorStaticSymbol = self.gen.ErrorStaticSymbol
        self.ErrorSet = self.gen.ErrorSet
        self.ErrorSetTwoQuditTensorProduct = self.gen.ErrorSetTwoQuditTensorProduct

    def start(
        self,
        circuit: QuantumCircuit,
        error_params: ErrorParams,
    ) -> ErrorCircuits:
        """Start up at the beginning of a circuit simulation.

        Args:
        ----
            circuit: The quantum circuit to simulate.
            error_params: Parameters controlling error generation.

        """
        self.error_circuits = ErrorCircuits()
        self.circuit = circuit
        self.error_params = error_params

        return self.error_circuits

    def generate_tick_errors(
        self,
        tick_circuit: QuantumCircuit,
        time: int | tuple[int, ...],
        **params: GateParams,
    ) -> ErrorCircuits:
        """Returns before errors, after errors, and replaced locations for the given key (args).

        Args:
        ----
            tick_circuit: The tick circuit containing gate operations.
            time: The time index or tuple indicating when errors occur.
            **params: Additional parameters including data_qudit_set.

        """
        tick_index = time[-1] if isinstance(time, tuple) else time

        circuit = tick_circuit.circuit

        # Simple model where for each gate there is a probability "p" for an X, Y, or Z error to occur.

        before = QuantumCircuit()
        after = QuantumCircuit()
        replace = set()

        # Data errors
        # -----------
        if tick_index == 0 and "data" in self.gen.error_func_dict:
            data_qudit_set = params["data_qudit_set"]
            self.gen.create_errors(self, "data", data_qudit_set, after, before, replace)

        # unitary and measurement errors
        # ------------------------------
        for symbol, gate_locations, _ in circuit.items(tick_index):
            self.gen.create_errors(self, symbol, gate_locations, after, before, replace)

        # idle errors
        # -----------
        if "idle" in self.gen.error_func_dict:
            inactive_qudits = circuit.qudits - circuit.active_qudits[tick_index]

            if tick_index == 0 and "data" in self.gen.error_func_dict:
                data_qudit_set = params["data_qudit_set"]
                inactive_qudits -= data_qudit_set

            self.gen.create_errors(
                self,
                "idle",
                inactive_qudits,
                after,
                before,
                replace,
            )

        self.error_circuits.add_circuits(time, before, after)

        return self.error_circuits

    def get_gate_error(
        self,
        symbol: str,
        gate_locations: LocationSet,
        error_params: ErrorParams,
    ) -> tuple[QuantumCircuit, QuantumCircuit, set]:
        """Get error for a specific gate operation.

        Args:
        ----
            symbol: The gate symbol/operation name.
            gate_locations: Locations where the gate is applied.
            error_params: Parameters controlling error generation.

        """
        self.error_params = error_params

        before = QuantumCircuit()
        after = QuantumCircuit()
        replace = set()
        self.gen.create_errors(self, symbol, gate_locations, after, before, replace)
        return after, before, replace



"""Test the unified sim API with different program types."""

import json

import pytest
from guppylang import guppy
from guppylang.std.quantum import h, measure, qubit
from pecos import sim
from pecos_rslib import (
    Hugr,
    PhirJson,
    Qasm,
    Qis,
    sparse_stabilizer,
    state_vector,
)


class TestQASMSimulation:
    """Test sim API with QASM programs."""

    def test_sim_api_with_simple_qasm(self) -> None:
        """Test sim API with simple QASM program."""
        qasm_str = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        creg c[1];
        h q[0];
        measure q[0] -> c[0];
        """

        program = Qasm.from_string(qasm_str)
        shot_vec = sim(program).seed(42).run(1000)

        # Convert ShotVec to dict
        results = shot_vec.to_dict()

        # Results is a dict with register names as keys, values are shot arrays
        assert isinstance(results, dict), "Results should be a dictionary"
        assert "c" in results, "Results should contain register 'c'"
        assert len(results["c"]) == 1000, "Should have 1000 shots"

        # Check measurement distribution (should be roughly 50/50)
        measurements = results["c"]
        ones = sum(measurements)
        zeros = 1000 - ones

        # With seed, results should be deterministic but still mixed
        assert 300 < ones < 700, f"Should be roughly 50/50 distribution, got {ones} ones"
        assert 300 < zeros < 700, f"Should be roughly 50/50 distribution, got {zeros} zeros"

    def test_sim_api_with_bell_state_qasm(self) -> None:
        """Test sim API with Bell state in QASM."""
        qasm_str = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        creg c[2];
        h q[0];
        cx q[0], q[1];
        measure q[0] -> c[0];
        measure q[1] -> c[1];
        """

        program = Qasm.from_string(qasm_str)
        shot_vec = sim(program).seed(42).run(100)
        results = shot_vec.to_dict()

        assert "c" in results, "Results should contain register 'c'"
        assert len(results["c"]) == 100, "Should have 100 shots"

        # Each shot should be a 2-bit value (0, 1, 2, or 3)
        # For Bell state, should only see 00 (0) and 11 (3)
        measurements = results["c"]
        unique_values = set(measurements)

        # Bell state should only produce correlated results
        assert unique_values.issubset(
            {0, 3},
        ), f"Bell state should only give 00 or 11, got {unique_values}"

        # Should see both values with reasonable probability
        count_00 = measurements.count(0)
        count_11 = measurements.count(3)
        assert count_00 > 20, f"Should see |00⟩ state, got {count_00} times"
        assert count_11 > 20, f"Should see |11⟩ state, got {count_11} times"

    def test_sim_builder_chaining(self) -> None:
        """Test builder pattern chaining."""
        qasm_str = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        creg c[1];
        x q[0];
        measure q[0] -> c[0];
        """

        program = Qasm.from_string(qasm_str)

        # Test chaining various configurations
        shot_vec = sim(program).seed(42).workers(2).quantum(state_vector()).run(500)
        results = shot_vec.to_dict()

        assert "c" in results, "Results should contain register 'c'"
        assert len(results["c"]) == 500, "Should have 500 shots"

        # X gate should always give |1⟩
        measurements = results["c"]
        assert all(m == 1 for m in measurements), "X gate should always measure 1"


class TestLLVMSimulation:
    """Test sim API with LLVM IR programs."""

    def test_sim_api_with_llvm_simple(self) -> None:
        """Test sim API with simple LLVM IR program."""
        # QIS format LLVM IR - uses i64 for qubit indices and qmain entry point
        # This matches the format that PECOS HUGR compiler generates
        llvm_ir = """
        ; ModuleID = 'quantum_test'
        source_filename = "quantum_test"

        @str_c = constant [2 x i8] c"c\\00"

        ; Entry point using qmain(i64)->i64 protocol
        define i64 @qmain(i64 %0) #0 {
        entry:
            ; Allocate qubit (returns i64)
            %qubit = call i64 @__quantum__rt__qubit_allocate()

            ; Apply H gate (takes i64 qubit index)
            call void @__quantum__qis__h__body(i64 %qubit)

            ; Allocate result handle
            %result_id = call i64 @__quantum__rt__result_allocate()

            ; Measure qubit (returns i32 measurement)
            %measurement = call i32 @__quantum__qis__m__body(i64 %qubit, i64 %result_id)

            ; Record output (convert i64 to i8* pointer for register name)
            %result_ptr = inttoptr i64 %result_id to i8*
            call void @__quantum__rt__result_record_output(i8* %result_ptr,
                i8* getelementptr inbounds ([2 x i8], [2 x i8]* @str_c, i32 0, i32 0))

            ret i64 0
        }

        declare i64 @__quantum__rt__qubit_allocate()
        declare void @__quantum__qis__h__body(i64)
        declare i64 @__quantum__rt__result_allocate()
        declare i32 @__quantum__qis__m__body(i64, i64)
        declare void @__quantum__rt__result_record_output(i8*, i8*)

        attributes #0 = { "EntryPoint" }
        """

        try:
            program = Qis.from_string(llvm_ir)

            # Try to run - this might work now with proper QIR format
            shot_vec = sim(program).qubits(1).seed(42).run(10)
            results = shot_vec.to_dict()

            # If it works, verify results
            assert isinstance(results, dict), "Results should be a dictionary"
            assert len(results) > 0, "Should have some results"

            # Check for measurements
            if "c" in results:
                measurements = results["c"]
                assert len(measurements) == 10, "Should have 10 shots"
                assert all(m in [0, 1] for m in measurements), "Measurements should be binary"

        except (RuntimeError, ValueError, NotImplementedError) as e:
            # Known LLVM runtime issues
            error_msg = str(e).lower()
            if any(
                x in error_msg
                for x in [
                    "entry",
                    "not implemented",
                    "undefined symbol",
                    "failed to load",
                ]
            ):
                pytest.skip(f"LLVM runtime not fully working yet: {e}")
            else:
                # Truly unexpected error
                pytest.fail(f"Unexpected LLVM simulation error: {e}")

    def test_sim_api_with_llvm_bell_state(self) -> None:
        """Test sim API with Bell state in LLVM IR."""
        # QIS format LLVM IR - uses i64 for qubit indices and qmain entry point
        # This matches the format that PECOS HUGR compiler generates
        llvm_ir = """
        ; ModuleID = 'bell_state'
        source_filename = "bell_state"

        @str_c0 = constant [3 x i8] c"c0\\00"
        @str_c1 = constant [3 x i8] c"c1\\00"

        ; Entry point using qmain(i64)->i64 protocol
        define i64 @qmain(i64 %0) #0 {
        entry:
            ; Allocate qubits (returns i64)
            %q0 = call i64 @__quantum__rt__qubit_allocate()
            %q1 = call i64 @__quantum__rt__qubit_allocate()

            ; Apply H to qubit 0
            call void @__quantum__qis__h__body(i64 %q0)

            ; Apply CX(q0, q1)
            call void @__quantum__qis__cx__body(i64 %q0, i64 %q1)

            ; Measure qubit 0
            %result_id0 = call i64 @__quantum__rt__result_allocate()
            %measurement0 = call i32 @__quantum__qis__m__body(i64 %q0, i64 %result_id0)
            %result_ptr0 = inttoptr i64 %result_id0 to i8*
            call void @__quantum__rt__result_record_output(i8* %result_ptr0,
                i8* getelementptr inbounds ([3 x i8], [3 x i8]* @str_c0, i32 0, i32 0))

            ; Measure qubit 1
            %result_id1 = call i64 @__quantum__rt__result_allocate()
            %measurement1 = call i32 @__quantum__qis__m__body(i64 %q1, i64 %result_id1)
            %result_ptr1 = inttoptr i64 %result_id1 to i8*
            call void @__quantum__rt__result_record_output(i8* %result_ptr1,
                i8* getelementptr inbounds ([3 x i8], [3 x i8]* @str_c1, i32 0, i32 0))

            ret i64 0
        }

        declare i64 @__quantum__rt__qubit_allocate()
        declare void @__quantum__qis__h__body(i64)
        declare void @__quantum__qis__cx__body(i64, i64)
        declare i64 @__quantum__rt__result_allocate()
        declare i32 @__quantum__qis__m__body(i64, i64)
        declare void @__quantum__rt__result_record_output(i8*, i8*)

        attributes #0 = { "EntryPoint" }
        """

        try:
            program = Qis.from_string(llvm_ir)
            shot_vec = sim(program).qubits(2).seed(42).run(50)
            results = shot_vec.to_dict()

            assert isinstance(results, dict), "Results should be a dictionary"

            # Check if we have correlated measurements
            if "c0" in results and "c1" in results:
                m0 = results["c0"]
                m1 = results["c1"]

                assert len(m0) == 50, "Should have 50 shots for qubit 0"
                assert len(m1) == 50, "Should have 50 shots for qubit 1"

                # Bell state should be correlated
                correlated = sum(1 for i in range(50) if m0[i] == m1[i])
                assert correlated == 50, f"Bell state should be perfectly correlated, got {correlated}/50"

        except (RuntimeError, ValueError, NotImplementedError) as e:
            error_msg = str(e).lower()
            if any(
                x in error_msg
                for x in [
                    "not implemented",
                    "not supported",
                    "undefined symbol",
                    "failed to load",
                    "getelementptr",
                    "unsized type",
                ]
            ):
                pytest.skip(f"LLVM Bell state not fully working yet: {e}")
            else:
                pytest.fail(f"Unexpected error: {e}")


@pytest.mark.optional_dependency
class TestHUGRSimulation:
    """Test sim API with HUGR programs."""

    def test_sim_api_with_real_hugr(self) -> None:
        """Test sim API with real HUGR from Guppy compilation."""

        # Create a real HUGR program from Guppy
        @guppy
        def simple_circuit() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        # Compile to HUGR
        compiled = simple_circuit.compile()

        # Get HUGR bytes - preferring to_bytes() which gives the correct format
        if hasattr(compiled, "to_bytes"):
            hugr_bytes = compiled.to_bytes()
        else:
            # Use to_str() for HUGR envelope format (includes header)
            hugr_str = compiled.to_str()
            hugr_bytes = hugr_str.encode("utf-8")

        try:
            program = Hugr.from_bytes(hugr_bytes)

            # This should route through Selene with HUGR 0.13
            shot_vec = sim(program).qubits(1).quantum(state_vector()).seed(42).run(100)
            results = shot_vec.to_dict()

            # If it works, verify results
            assert isinstance(results, dict), "Results should be a dictionary"

            # Check for measurements
            has_measurements = "measurement_1" in results or "measurements" in results or len(results) > 0
            assert has_measurements, "Should have measurement results"

            if "measurement_1" in results:
                measurements = results["measurement_1"]
                assert len(measurements) == 100, "Should have 100 measurements"

                # Should be roughly 50/50 for H gate
                ones = sum(measurements)
                assert 30 < ones < 70, f"H gate should give roughly 50/50, got {ones}/100"

        except (
            ImportError,
            RuntimeError,
            ValueError,
            NotImplementedError,
            TypeError,
        ) as e:
            error_msg = str(e).lower()
            if "hugr" in error_msg and "not implemented" in error_msg:
                pytest.skip(f"HUGR parsing not fully implemented: {e}")
            elif "not supported" in error_msg:
                pytest.skip(f"HUGR not fully supported: {e}")
            elif "unknown resource type" in error_msg and "hugrprogram" in error_msg:
                pytest.skip(f"Hugr type not properly recognized by sim API: {e}")
            else:
                # This might be a real error worth investigating
                pytest.fail(f"Unexpected HUGR simulation error: {e}")

    def test_sim_api_hugr_routing(self) -> None:
        """Test that HUGR programs route through compilation to Selene engine."""

        # Create a real HUGR program from Guppy for routing test
        @guppy
        def simple_h_measure() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        # Compile to HUGR
        compiled = simple_h_measure.compile()

        # Get HUGR bytes
        if hasattr(compiled, "to_bytes"):
            hugr_bytes = compiled.to_bytes()
        else:
            hugr_str = compiled.to_str()
            hugr_bytes = hugr_str.encode("utf-8")

        try:
            program = Hugr.from_bytes(hugr_bytes)

            # Create builder - this should work with real HUGR
            builder = sim(program)
            assert builder is not None, "Should create sim builder for HUGR"

            # Builder should have the right methods
            assert hasattr(builder, "qubits"), "Builder should have qubits method"
            assert hasattr(builder, "run"), "Builder should have run method"
            assert hasattr(builder, "quantum"), "Builder should have quantum method"

            # Configure and verify builder works
            configured = builder.qubits(1).quantum(state_vector())
            assert configured is not None, "Should configure builder"

        except (ImportError, RuntimeError) as e:
            error_msg = str(e).lower()
            if "selene" in error_msg:
                pytest.skip("Selene not available for HUGR routing")
            elif "hugr" in error_msg and "not implemented" in error_msg:
                pytest.skip(f"HUGR compilation not fully implemented: {e}")
            else:
                pytest.fail(f"Unexpected error in HUGR routing: {e}")


class TestPHIRSimulation:
    """Test sim API with PHIR JSON programs."""

    def test_sim_api_with_phir_basic(self) -> None:
        """Test sim API with basic PHIR JSON program."""
        # PHIR format for simple H gate and measurement
        phir_json = {
            "format": "PHIR/JSON",
            "version": "0.1.0",
            "metadata": {"description": "Simple H gate test"},
            "ops": [
                {
                    "data": "qvar_define",
                    "data_type": "qubits",
                    "variable": "q",
                    "size": 1,
                },
                {
                    "data": "cvar_define",
                    "data_type": "i64",
                    "variable": "m",
                    "size": 1,
                },
                {"qop": "H", "args": [["q", 0]]},
                {"qop": "Measure", "args": [["q", 0]], "returns": [["m", 0]]},
                {"cop": "Result", "args": ["m"], "returns": ["c"]},
            ],
        }

        phir_str = json.dumps(phir_json)

        program = PhirJson.from_string(phir_str)
        shot_vec = sim(program).qubits(1).seed(42).run(50)
        results = shot_vec.to_dict()

        assert isinstance(results, dict), "Results should be a dictionary"
        assert "c" in results, "Results should contain register 'c'"

        measurements = results["c"]
        assert len(measurements) == 50, "Should have 50 measurements"

        # Should be binary values
        assert all(m in [0, 1] for m in measurements), "Measurements should be binary"

        # H gate should give roughly 50/50 distribution
        ones = sum(measurements)
        assert 15 < ones < 35, f"H gate should give roughly 50/50, got {ones}/50"

    def test_sim_api_with_phir_bell_state(self) -> None:
        """Test sim API with Bell state in PHIR format."""
        phir_json = {
            "format": "PHIR/JSON",
            "version": "0.1.0",
            "metadata": {"description": "Bell state"},
            "ops": [
                {
                    "data": "qvar_define",
                    "data_type": "qubits",
                    "variable": "q",
                    "size": 2,
                },
                {
                    "data": "cvar_define",
                    "data_type": "i64",
                    "variable": "m",
                    "size": 2,
                },
                {"qop": "H", "args": [["q", 0]]},
                {"qop": "CX", "args": [["q", 0], ["q", 1]]},
                {"qop": "Measure", "args": [["q", 0]], "returns": [["m", 0]]},
                {"qop": "Measure", "args": [["q", 1]], "returns": [["m", 1]]},
                {"cop": "Result", "args": ["m"], "returns": ["c"]},
            ],
        }

        phir_str = json.dumps(phir_json)

        program = PhirJson.from_string(phir_str)
        shot_vec = sim(program).qubits(2).seed(42).run(100)
        results = shot_vec.to_dict()

        assert isinstance(results, dict), "Results should be a dictionary"
        assert "c" in results, "Results should contain register 'c'"

        measurements = results["c"]
        assert len(measurements) == 100, "Should have 100 measurements"

        # Bell state should only produce 00 (0) and 11 (3) in 2-bit encoding
        unique_values = set(measurements)
        assert unique_values.issubset(
            {0, 3},
        ), f"Bell state should only give 00 or 11, got {unique_values}"

        # Should see both values with reasonable probability
        count_00 = measurements.count(0)
        count_11 = measurements.count(3)
        assert count_00 > 20, f"Should see |00⟩ state, got {count_00} times"
        assert count_11 > 20, f"Should see |11⟩ state, got {count_11} times"


class TestSimAPIFeatures:
    """Test various features of the sim API."""

    def test_sim_with_different_backends(self) -> None:
        """Test sim API with different quantum backends."""
        qasm_str = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        creg c[1];
        h q[0];
        measure q[0] -> c[0];
        """

        program = Qasm.from_string(qasm_str)

        # Test with state vector backend
        shot_vec_sv = sim(program).quantum(state_vector()).seed(42).run(100)
        results_sv = shot_vec_sv.to_dict()
        assert "c" in results_sv, "State vector backend should produce results"

        # Test with sparse stabilizer backend
        try:
            shot_vec_ss = sim(program).quantum(sparse_stabilizer()).seed(42).run(100)
            results_ss = shot_vec_ss.to_dict()
            assert "c" in results_ss, "Sparse stabilizer backend should produce results"

            # Results might differ between backends but both should be valid
            assert len(results_sv["c"]) == 100, "State vector should give 100 shots"
            assert len(results_ss["c"]) == 100, "Sparse stabilizer should give 100 shots"

        except (RuntimeError, ValueError) as e:
            if "not supported" in str(e).lower():
                pytest.skip(f"Sparse stabilizer not supported for this program: {e}")

    def test_sim_error_handling(self) -> None:
        """Test error handling in sim API."""
        # Invalid QASM
        invalid_qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        invalid_gate q[0];
        """

        program = Qasm.from_string(invalid_qasm)
        with pytest.raises((RuntimeError, ValueError)) as exc_info:
            sim(program).run(10)

        assert (
            "invalid" in str(exc_info.value).lower() or "error" in str(exc_info.value).lower()
        ), "Should raise error for invalid QASM"

    def test_sim_deterministic_seeding(self) -> None:
        """Test that seeding produces deterministic results."""
        qasm_str = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        creg c[2];
        h q[0];
        h q[1];
        measure q[0] -> c[0];
        measure q[1] -> c[1];
        """

        program = Qasm.from_string(qasm_str)

        # Run twice with same seed
        shot_vec1 = sim(program).seed(12345).run(50)
        shot_vec2 = sim(program).seed(12345).run(50)
        results1 = shot_vec1.to_dict()
        results2 = shot_vec2.to_dict()

        assert "c" in results1, "First run should produce results"
        assert "c" in results2, "Second run should produce results"

        # Results should be identical with same seed
        assert results1["c"] == results2["c"], "Same seed should give identical results"

        # Run with different seed
        shot_vec3 = sim(program).seed(54321).run(50)
        results3 = shot_vec3.to_dict()

        # Results should differ with different seed (statistically)
        assert results1["c"] != results3["c"], "Different seeds should give different results"



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""Magic State Distillation protocol geometry.

This module provides geometry structures for magic state distillation (MSD)
protocols using surface codes. MSD uses a two-level approach:

1. Inner code (distance-2): Fast error detection with 4 data qubits
2. Outer code (distance-3): Full surface code protection with 9 data qubits

The inner code acts as a filter - if errors are detected, the state is
discarded before expensive outer code expansion.

Qubit layout (3x3 grid):
    0  1  2
    3  4  5
    6  7  8

Inner code uses qubits {0, 1, 3, 4} (top-left 2x2).
Outer code adds qubits {2, 5, 6, 7, 8} (right column and bottom row).
"""

from dataclasses import dataclass, field


@dataclass(frozen=True)
class InnerCodeGeometry:
    """Distance-2 inner code geometry for MSD.

    The inner code uses 4 data qubits in a 2x2 arrangement from the
    top-left corner of the full 3x3 grid.

    Attributes:
        data_qubits: Indices of data qubits (0, 1, 3, 4).
        z_stabilizer: Data qubit indices for the Z stabilizer.
        x_stabilizers: Data qubit indices for X stabilizers (top, bottom).
    """

    data_qubits: tuple[int, ...] = (0, 1, 3, 4)

    # Z stabilizer: measures Z on all 4 qubits (column-major order)
    z_stabilizer: tuple[int, ...] = (0, 3, 1, 4)

    # X stabilizers: top row [0,1] and bottom row [3,4]
    x_stabilizers: tuple[tuple[int, ...], ...] = ((0, 1), (3, 4))

    @property
    def num_data(self) -> int:
        """Number of data qubits."""
        return len(self.data_qubits)

    @property
    def num_x_stabilizers(self) -> int:
        """Number of X stabilizers."""
        return len(self.x_stabilizers)

    @property
    def num_z_stabilizers(self) -> int:
        """Number of Z stabilizers."""
        return 1

    @property
    def num_syndromes(self) -> int:
        """Total syndrome bits per round."""
        return self.num_x_stabilizers + self.num_z_stabilizers


@dataclass(frozen=True)
class OuterCodeGeometry:
    """Distance-3 outer code geometry for MSD.

    The outer code is a full d=3 surface code with 9 data qubits.

    Attributes:
        data_qubits: Indices of all 9 data qubits.
        inner_qubits: Indices that come from inner code (0, 1, 3, 4).
        expansion_qubits: New qubits added for outer code (2, 5, 6, 7, 8).
        x_stabilizers: X stabilizer supports (4 total: 2 bulk, 2 boundary).
        z_stabilizers: Z stabilizer supports (4 total: 2 bulk, 2 boundary).
    """

    data_qubits: tuple[int, ...] = (0, 1, 2, 3, 4, 5, 6, 7, 8)
    inner_qubits: tuple[int, ...] = (0, 1, 3, 4)
    expansion_qubits: tuple[int, ...] = (2, 5, 6, 7, 8)

    # X stabilizers (4 total)
    # Boundary: [0,1] (top), [7,8] (bottom)
    # Bulk: [1,2,4,5], [3,4,6,7]
    x_stabilizers: tuple[tuple[int, ...], ...] = (
        (0, 1),  # boundary top
        (1, 2, 4, 5),  # bulk
        (3, 4, 6, 7),  # bulk
        (7, 8),  # boundary bottom
    )

    # Z stabilizers (4 total)
    # Boundary: [2,5] (right), [3,6] (left)
    # Bulk: [0,1,3,4] (same as inner Z), [4,5,7,8]
    z_stabilizers: tuple[tuple[int, ...], ...] = (
        (2, 5),  # boundary right
        (0, 1, 3, 4),  # bulk (inner code stabilizer)
        (4, 5, 7, 8),  # bulk
        (3, 6),  # boundary left
    )

    @property
    def num_data(self) -> int:
        """Number of data qubits."""
        return len(self.data_qubits)

    @property
    def num_expansion(self) -> int:
        """Number of qubits added during expansion."""
        return len(self.expansion_qubits)

    @property
    def num_x_stabilizers(self) -> int:
        """Number of X stabilizers."""
        return len(self.x_stabilizers)

    @property
    def num_z_stabilizers(self) -> int:
        """Number of Z stabilizers."""
        return len(self.z_stabilizers)

    @property
    def num_syndromes(self) -> int:
        """Total syndrome bits per round."""
        return self.num_x_stabilizers + self.num_z_stabilizers

    def get_bulk_x_stabilizers(self) -> tuple[tuple[int, ...], ...]:
        """Get bulk (weight-4) X stabilizers."""
        return tuple(s for s in self.x_stabilizers if len(s) == 4)

    def get_boundary_x_stabilizers(self) -> tuple[tuple[int, ...], ...]:
        """Get boundary (weight-2) X stabilizers."""
        return tuple(s for s in self.x_stabilizers if len(s) == 2)

    def get_bulk_z_stabilizers(self) -> tuple[tuple[int, ...], ...]:
        """Get bulk (weight-4) Z stabilizers."""
        return tuple(s for s in self.z_stabilizers if len(s) == 4)

    def get_boundary_z_stabilizers(self) -> tuple[tuple[int, ...], ...]:
        """Get boundary (weight-2) Z stabilizers."""
        return tuple(s for s in self.z_stabilizers if len(s) == 2)


@dataclass(frozen=True)
class MSDProtocol:
    """Magic State Distillation protocol geometry.

    Combines inner and outer code geometry with protocol parameters.

    Attributes:
        inner: Inner code geometry (distance-2).
        outer: Outer code geometry (distance-3).
        inner_rounds: Number of syndrome extraction rounds for inner code.
        outer_rounds: Number of syndrome extraction rounds for outer code.
    """

    inner: InnerCodeGeometry = field(default_factory=InnerCodeGeometry)
    outer: OuterCodeGeometry = field(default_factory=OuterCodeGeometry)
    inner_rounds: int = 2
    outer_rounds: int = 1

    @property
    def total_data_qubits(self) -> int:
        """Total data qubits needed (same as outer code)."""
        return self.outer.num_data

    @property
    def inner_syndrome_bits(self) -> int:
        """Syndrome bits per inner code round."""
        return self.inner.num_syndromes

    @property
    def outer_syndrome_bits(self) -> int:
        """Syndrome bits per outer code round."""
        return self.outer.num_syndromes

    @property
    def total_inner_syndromes(self) -> int:
        """Total syndrome bits from inner code (all rounds)."""
        return self.inner.num_syndromes * self.inner_rounds

    @property
    def total_outer_syndromes(self) -> int:
        """Total syndrome bits from outer code (all rounds)."""
        return self.outer.num_syndromes * self.outer_rounds

    def get_expansion_prep_states(self) -> dict[int, str]:
        """Get preparation states for expansion qubits.

        When expanding from inner to outer code:
        - Bottom row qubits (6, 7, 8) are prepared in |+>
        - Right column qubits (2, 5) are prepared in |0>

        Returns:
            Dict mapping qubit index to preparation state ('+' or '0').
        """
        return {
            2: "0",  # Right column: |0>
            5: "0",  # Right column: |0>
            6: "+",  # Bottom row: |+>
            7: "+",  # Bottom row: |+>
            8: "+",  # Bottom row: |+>
        }

    def get_inner_init_states(self) -> dict[int, str]:
        """Get initialization states for inner code qubits.

        For T state distillation:
        - Qubit 0: T|+> (T gate applied to |+>)
        - Qubit 1: |0>
        - Qubits 3, 4: |+>

        Returns:
            Dict mapping qubit index to preparation state.
        """
        return {
            0: "T+",  # T|+> magic state
            1: "0",  # |0>
            3: "+",  # |+>
            4: "+",  # |+>
        }


def create_msd_protocol(
    inner_rounds: int = 2,
    outer_rounds: int = 1,
) -> MSDProtocol:
    """Create an MSD protocol with specified parameters.

    Args:
        inner_rounds: Number of syndrome extraction rounds for inner code.
            Default is 2 (two rounds, check for consistency).
        outer_rounds: Number of syndrome extraction rounds for outer code.
            Default is 1.

    Returns:
        Configured MSDProtocol instance.
    """
    return MSDProtocol(
        inner=InnerCodeGeometry(),
        outer=OuterCodeGeometry(),
        inner_rounds=inner_rounds,
        outer_rounds=outer_rounds,
    )



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Test configuration and shared fixtures."""

# Configure matplotlib to use non-interactive backend for tests (if available)
# This must be done before importing matplotlib.pyplot to avoid GUI backend issues on Windows
try:
    import matplotlib as mpl

    mpl.use("Agg")
except ImportError:
    # matplotlib is optional - only needed for visualization tests
    pass

# Note: llvmlite functionality is now always available via Rust (pecos_rslib.ir and pecos_rslib.binding)
# No need for conditional test skipping



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Type-safe list implementation with runtime type checking.

This module provides a TypedList class that enforces type constraints on list
elements at runtime, ensuring all items conform to a specified type throughout
the list's lifetime.
"""

from __future__ import annotations

from collections.abc import Iterable
from typing import Generic, TypeVar

T = TypeVar("T")


class TypedList(list[T], Generic[T]):
    """A list that only accepts items of a specific type, where are checked at runtime.

    Attributes:
        type (Type[T]): The type of the elements that the list should contain.

    Methods:
        append(item): Appends an item to the end of the list, checking its type.
        extend(iterable): Extends the list by appending elements from the iterable, checking their types.
        insert(index, item): Inserts an item at a given position, checking its type.
        __setitem__(index, item): Sets the item at the specified position or slice, checking its type.
    """

    def __init__(self, _type: type, data: Iterable[T] | None = None) -> None:
        """Initializes a TypedList with as specified element type and optional initial data.

        Args:
            _type (Type[T]): The type of the elements that the list should contain.
            data (Iterable[T], optional): Initial data to populate the typed list. Defaults to None.

        Raises:
             TypeError: If any item in the initial data is not of the correct type.
        """
        self.type = _type
        if data is not None:
            for i in data:
                self._check_type(i)
            super().__init__(data)
        else:
            super().__init__()

    def _check_type(self, /, __object: T) -> None:
        """Checks if an item is of the specified type.

        Args:
            __object (T): The item to check.

        Raises:
            TypeError: If the item is not of the correct type.
        """
        if not isinstance(__object, self.type):
            msg = f"Item must be of type {self.type.__name__}, got type {type(__object).__name__} instead"
            raise TypeError(msg)

    def append(self, /, __object: T) -> None:
        """Appends an item to the end of the list, checking its type.

        Args:
            __object (T): The item to append.

        Raises:
            TypeError: If the item is not of the correct type.
        """
        self._check_type(__object)
        super().append(__object)

    def extend(self, /, __iterable: Iterable[T]) -> None:
        """Extends the list by appending elements from the iterable, checking their type.

        Args:
            __iterable (Iterable[T]): An iterable of items to extend the list with.

        Raises:
            TypeError: If any item in the iterable is not of the correct type.
        """
        for i in __iterable:
            self._check_type(i)
        super().extend(__iterable)

    def insert(self, /, __index: int, __object: T) -> None:
        """Inserts an item en position, checking its type.

        Args:
            __index (int): The pto insert the .
            __object (T): The item to insert.
        """
        self._check_type(__object)
        super().insert(__index, __object)

    def __setitem__(self, key: int | slice, value: T | Iterable[T]) -> None:
        """Sets the item at the specified position or slice, checking its type.

        Args:
            key (int | slice): The position or slice to set the item(s) at.
            value (T | Iterable[T]): The item(s) to set.

        Raises:
            TypeError: If the item or any item in the slice is not of the correct type.
        """
        if isinstance(value, slice):
            if isinstance(value, Iterable):
                for i in value:
                    self._check_type(i)
            else:
                self._check_type(value)
        else:
            self._check_type(value)
        super().__setitem__(key, value)



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Steane code implementation with logical operations and error correction.

This module provides a complete implementation of the Steane [[7,1,3]] quantum error correction code,
including state preparation, logical gates, measurements, and quantum error correction protocols.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

from pecos.slr import Block, CReg, If, Permute, QReg, Vars

# from pecos.slr.qeclib.qubit.qubit import PhysicalQubit
# TODO: Use physical qubit implementation... accept multiple arguments
# TODO: Make sure the Steane gate api matches the physical qubit api
# TODO: Consider using Protocol to avoid inheritance but ensure unified API
from pecos.slr.qeclib.steane.gates_sq import paulis, sqrt_paulis
from pecos.slr.qeclib.steane.gates_sq.hadamards import H
from pecos.slr.qeclib.steane.gates_tq import transversal_tq
from pecos.slr.qeclib.steane.meas.destructive_meas import MeasDecode
from pecos.slr.qeclib.steane.preps.pauli_states import PrepRUS
from pecos.slr.qeclib.steane.preps.t_plus_state import (
    PrepEncodeTPlusFTRUS,
    PrepEncodeTPlusNonFT,
)
from pecos.slr.qeclib.steane.qec.qec_3parallel import ParallelFlagQECActiveCorrection

if TYPE_CHECKING:
    from pecos.slr import Bit


class Steane(Vars):
    """A generic implementation of a Steane code and operations.

    This represents one particular choice of Steane protocols. For finer control construct your own class
    or utilize the library of Steane code protocols directly.
    """

    def __init__(
        self,
        name: str,
        default_rus_limit: int = 3,
        ancillas: QReg | None = None,
    ) -> None:
        """Initialize a Steane logical qubit with data and ancilla registers.

        Args:
            name: Name prefix for all registers associated with this logical qubit.
            default_rus_limit: Default limit for repeat-until-success protocols.
            ancillas: Optional pre-allocated ancilla register (must have >= 3 qubits).
        """
        super().__init__()
        self.d = QReg(f"{name}_d", 7)
        self.a = ancillas or QReg(f"{name}_a", 3)
        self.c = CReg(f"{name}_c", 32)

        if self.a.size < 3:
            msg = f"Steane ancilla registers must have >= 3 qubits (provided: {self.a.size})"
            raise ValueError(msg)

        # TODO: Make it so I can put these in self.c... need to convert things like if(c) and c = a ^ b, a = 0;
        #  to allow lists of bits
        self.syn_meas = CReg(f"{name}_syn_meas", 32)
        self.last_raw_syn_x = CReg(f"{name}_last_raw_syn_x", 32)
        self.last_raw_syn_z = CReg(f"{name}_last_raw_syn_z", 32)
        self.scratch = CReg(f"{name}_scratch", 32)
        self.flag_x = CReg(f"{name}_flag_x", 3)
        self.flag_z = CReg(f"{name}_flags_z", 3)

        self.flags = CReg(f"{name}_flags", 3)  # weird error when using [c, c, c]

        self.raw_meas = CReg(f"{name}_raw_meas", 7)

        self.syn_x = CReg(f"{name}_syn_x", 3)
        self.syn_z = CReg(f"{name}_syn_z", 3)
        self.syndromes = CReg(f"{name}_syndromes", 3)
        self.verify_prep = CReg(f"{name}_verify_prep", 32)

        self.vars = [
            self.d,
        ]

        if ancillas is None:
            self.vars.append(self.a)

        self.vars.extend(
            [
                self.c,
                self.syn_meas,
                self.last_raw_syn_x,
                self.last_raw_syn_z,
                self.scratch,
                self.flag_x,
                self.flag_z,
                self.flags,
                self.raw_meas,
                self.syn_x,
                self.syn_z,
                self.syndromes,
                self.verify_prep,
            ],
        )

        # derived classical registers
        c = self.c
        self.log_raw = c[1]
        self.log = c[2]
        self.pf_x = c[3]
        self.pf_z = c[4]
        self.t_meas = c[5]
        self.tdg_meas = c[6]

        self.default_rus_limit = default_rus_limit

    @staticmethod
    def p(
        qubit: Steane,
        state: str,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare a logical qubit in a logical Pauli basis state."""
        block = PrepRUS(
            q=qubit.d,
            a=qubit.a[0],
            init=qubit.verify_prep[0],
            limit=rus_limit or qubit.default_rus_limit,
            state=state,
            first_round_reset=True,
        )
        if reject is not None:
            block.extend(reject.set(qubit.verify_prep[0]))
        return block

    @staticmethod
    def px(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical |+X>, a.k.a. |+>."""
        return qubit.p(qubit, "+X", reject=reject, rus_limit=rus_limit)

    @staticmethod
    def pnx(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical |-X>, a.k.a. |->."""
        return qubit.p(qubit, "-X", reject=reject, rus_limit=rus_limit)

    @staticmethod
    def py(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical |+Y>, a.k.a. |+i>."""
        return qubit.p(qubit, "+Y", reject=reject, rus_limit=rus_limit)

    @staticmethod
    def pny(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical |-Y>, a.k.a. |-i>."""
        return qubit.p(qubit, "-Y", reject=reject, rus_limit=rus_limit)

    @staticmethod
    def pz(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical |+Z>, a.k.a. |0>."""
        return qubit.p(qubit, "+Z", reject=reject, rus_limit=rus_limit)

    @staticmethod
    def pnz(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical |-Z>, a.k.a. |1>."""
        return qubit.p(qubit, "-Z", reject=reject, rus_limit=rus_limit)

    @classmethod
    def nonft_prep_t_plus_state(cls, qubit: Steane) -> Block:
        """Prepare logical T|+X> in a non-fault tolerant manner."""
        return PrepEncodeTPlusNonFT(
            q=qubit.d,
        )

    @staticmethod
    def prep_t_plus_state(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical T|+X> in a fault-tolerant manner."""
        block = Block(
            qubit.scratch.set(0),
            PrepEncodeTPlusFTRUS(
                d=qubit.d,
                a=qubit.a,
                out=qubit.scratch,
                reject=qubit.scratch[2],  # the first two bits of self.scratch are used by "out"
                flag_x=qubit.flag_x,
                flag_z=qubit.flag_z,
                flags=qubit.flags,
                last_raw_syn_x=qubit.last_raw_syn_x,
                last_raw_syn_z=qubit.last_raw_syn_z,
                limit=rus_limit or qubit.default_rus_limit,
            ),
        )
        if reject is not None:
            block.extend(reject.set(qubit.scratch[2]))
        return block

    @staticmethod
    def nonft_prep_tdg_plus_state(qubit: Steane) -> Block:
        """Prepare logical Tdg|+X> in a non-fault tolerant manner."""
        return Block(
            qubit.nonft_prep_t_plus_state(qubit),
            qubit.z(qubit),
        )

    @staticmethod
    def prep_tdg_plus_state(
        qubit: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical Tdg|+X> in a fault-tolerant manner."""
        return Block(
            qubit.prep_t_plus_state(qubit=qubit, reject=reject, rus_limit=rus_limit),
            qubit.szdg(qubit),
        )

    @staticmethod
    def x(qubit: Steane) -> Block:
        """Logical Pauli X gate."""
        return paulis.X(qubit.d)

    @staticmethod
    def y(qubit: Steane) -> Block:
        """Logical Pauli Y gate."""
        return paulis.Y(qubit.d)

    @staticmethod
    def z(qubit: Steane) -> Block:
        """Logical Pauli Z gate."""
        return paulis.Z(qubit.d)

    @staticmethod
    def h(qubit: Steane) -> Block:
        """Logical Hadamard gate."""
        return H(qubit.d)

    @staticmethod
    def sx(qubit: Steane) -> Block:
        """Sqrt of X."""
        return sqrt_paulis.SX(qubit.d)

    @staticmethod
    def sxdg(qubit: Steane) -> Block:
        """Adjoint of sqrt of X."""
        return sqrt_paulis.SXdg(qubit.d)

    @staticmethod
    def sy(qubit: Steane) -> Block:
        """Sqrt of Y."""
        return sqrt_paulis.SY(qubit.d)

    @staticmethod
    def sydg(qubit: Steane) -> Block:
        """Adjoint of sqrt of Y."""
        return sqrt_paulis.SYdg(qubit.d)

    @staticmethod
    def sz(qubit: Steane) -> Block:
        """Sqrt of Z. Also known as the S gate."""
        return sqrt_paulis.SZ(qubit.d)

    @staticmethod
    def szdg(qubit: Steane) -> Block:
        """Adjoint of sqrt of Z. Also known as the Sdg gate."""
        return sqrt_paulis.SZdg(qubit.d)

    @staticmethod
    def nonft_t(qubit: Steane, aux: Steane) -> Block:
        """T gate via teleportation using non-fault-tolerant initialization of the T|+> state."""
        return Block(
            qubit.nonft_prep_t_plus_state(aux),
            qubit.cx(qubit, aux),
            qubit.mz(qubit, qubit.t_meas),
            If(qubit.t_meas == 1).Then(qubit.sz(qubit)),
        )

    @staticmethod
    def t(
        qubit: Steane,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """T gate via teleportation using fault-tolerant initialization of the T|+> state."""
        return Block(
            qubit.prep_t_plus_state(aux, reject=reject, rus_limit=rus_limit),
            qubit.cx(qubit, aux),
            qubit.mz(qubit, qubit.t_meas),
            If(qubit.t_meas == 1).Then(qubit.sz(qubit)),  # SZ/S correction.
        )

    @staticmethod
    def nonft_tdg(qubit: Steane, aux: Steane) -> Block:
        """Tdg gate via teleportation using non-fault-tolerant initialization of the Tdg|+> state."""
        return Block(
            qubit.nonft_prep_tdg_plus_state(qubit),
            qubit.cx(qubit, aux),
            qubit.mz(qubit, qubit.tdg_meas),
            If(qubit.tdg_meas == 1).Then(qubit.szdg(qubit)),
        )

    @staticmethod
    def tdg(
        qubit: Steane,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Tdg gate via teleportation using fault-tolerant initialization of the Tdg|+> state."""
        return Block(
            qubit.prep_tdg_plus_state(qubit, reject=reject, rus_limit=rus_limit),
            qubit.cx(qubit, aux),
            qubit.mz(aux, aux.tdg_meas),
            If(qubit.tdg_meas == 1).Then(qubit.szdg(qubit)),  # SZdg/Sdg correction.
        )

    #  Begin Experimental: ------------------------------------
    @staticmethod
    def nonft_t_tel(qubit: Steane, aux: Steane) -> Block:
        """Warning: This is experimental.

        T gate via teleportation using non-fault-tolerant initialization of the T|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: nonft_t_tel", stacklevel=2)
        return Block(
            qubit.nonft_prep_t_plus_state(aux),
            qubit.cx(aux, qubit),
            qubit.mz(qubit.t_meas),
            If(qubit.t_meas == 1).Then(aux.x(aux), aux.sz(aux)),
            Permute(qubit.d, aux.d),
        )

    @staticmethod
    def t_tel(
        qubit: Steane,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Warning: This is experimental.

        T gate via teleportation using fault-tolerant initialization of the T|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: t_tel", stacklevel=2)
        return Block(
            aux.prep_t_plus_state(aux, reject=reject, rus_limit=rus_limit),
            aux.cx(aux, qubit),
            qubit.mz(qubit.t_meas),
            If(qubit.t_meas == 1).Then(aux.x(aux), aux.sz(aux)),  # SZ/S correction.
            Permute(qubit.d, aux.d),
        )

    @staticmethod
    def nonft_tdg_tel(qubit: Steane, aux: Steane) -> Block:
        """Warning: This is experimental.

        Tdg gate via teleportation using non-fault-tolerant initialization of the Tdg|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: nonft_tdg_tel", stacklevel=2)
        return Block(
            aux.nonft_prep_tdg_plus_state(aux),
            aux.cx(aux, qubit),
            qubit.mz(qubit, qubit.tdg_meas),
            If(qubit.tdg_meas == 1).Then(aux.x(aux), aux.szdg(aux)),
            Permute(qubit.d, aux.d),
        )

    @staticmethod
    def tdg_tel(
        qubit: Steane,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Warning: This is experimental.

        Tdg gate via teleportation using fault-tolerant initialization of the Tdg|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: tdg_tel", stacklevel=2)
        return Block(
            aux.prep_tdg_plus_state(aux, reject=reject, rus_limit=rus_limit),
            aux.cx(aux, qubit),
            qubit.mz(qubit, qubit.tdg_meas),
            If(qubit.t_meas == 1).Then(
                aux.x(aux),
                aux.szdg(aux),
            ),  # SZdg/Sdg correction.
            Permute(aux.d, aux.d),
        )

    # End Experimental: ------------------------------------

    @staticmethod
    def cx(qubit: Steane, target: Steane) -> Block:
        """Logical CX."""
        return transversal_tq.CX(qubit.d, target.d)

    @staticmethod
    def cy(qubit: Steane, target: Steane) -> Block:
        """Logical CY."""
        return transversal_tq.CY(qubit.d, target.d)

    @staticmethod
    def cz(qubit: Steane, target: Steane) -> Block:
        """Logical CZ."""
        return transversal_tq.CZ(qubit.d, target.d)

    @staticmethod
    def m(qubit: Steane, meas_basis: str, log: Bit | None = None) -> Block:
        """Destructively measure the logical qubit in some Pauli basis."""
        block = Block(
            MeasDecode(
                q=qubit.d,
                meas_basis=meas_basis,
                meas=qubit.raw_meas,
                log_raw=qubit.log_raw,
                log=qubit.log,
                syn_meas=qubit.syn_meas,
                pf_x=qubit.pf_x,
                pf_z=qubit.pf_z,
                last_raw_syn_x=qubit.last_raw_syn_x,
                last_raw_syn_z=qubit.last_raw_syn_z,
            ),
        )
        if log is not None:
            block.extend(log.set(qubit.log))
        return block

    @staticmethod
    def mx(qubit: Steane, log: Bit | None = None) -> Block:
        """Logical destructive measurement of the logical X operator."""
        return qubit.m("X", log=log)

    @staticmethod
    def my(qubit: Steane, log: Bit | None = None) -> Block:
        """Logical destructive measurement of the logical Y operator."""
        return qubit.m("Y", log=log)

    @staticmethod
    def mz(qubit: Steane, log: Bit | None = None) -> Block:
        """Logical destructive measurement of the logical Z operator."""
        return qubit.m("Z", log=log)

    def qec(self, flag_bit: Bit | None = None) -> Block:
        """Perform quantum error correction on the logical qubit.

        Runs parallel flag quantum error correction with active correction.

        Args:
            flag_bit: Optional bit to set if any flags are raised during QEC.

        Returns:
            Block containing the QEC operations.
        """
        block = ParallelFlagQECActiveCorrection(
            q=self.d,
            a=self.a,
            flag_x=self.flag_x,
            flag_z=self.flag_z,
            flags=self.flags,
            syn_x=self.syn_x,
            syn_z=self.syn_z,
            last_raw_syn_x=self.last_raw_syn_x,
            last_raw_syn_z=self.last_raw_syn_z,
            syndromes=self.syndromes,
            pf_x=self.pf_x,
            pf_z=self.pf_z,
            scratch=self.scratch,
        )
        if flag_bit is not None:
            block.extend(If(self.flags != 0).Then(flag_bit.set(1)))
        return block



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
Qubit Allocator for SLR.

Inspired by Zig's allocator pattern and NASA's Power of 10 rules.
Provides hierarchical qubit slot management with explicit lifecycle states.

See docs/proposals/slr-qubit-allocators.md for full design documentation.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator


class SlotState(Enum):
    """State of a qubit slot in an allocator.

    Two states only:
    - UNPREPARED: Not ready for gates (initial state or after measurement)
    - PREPARED: Ready for gate operations
    """

    UNPREPARED = "unprepared"
    PREPARED = "prepared"


class QubitRef:
    """A reference to a qubit slot in an allocator.

    Used as arguments to gate operations. Not a standalone qubit -
    always tied to its parent allocator.

    QubitRef is ephemeral - created for immediate gate application,
    not stored. Like Zig slices, valid only while allocator is alive.

    Provides compatibility with legacy Qubit interface:
    - `ref.reg` returns the allocator (like Qubit.reg returns QReg)
    - `ref.index` returns the slot index
    """

    __slots__ = ("_alloc", "_index")

    def __init__(self, alloc: QAlloc, index: int) -> None:
        self._alloc = alloc
        self._index = index

    @property
    def alloc(self) -> QAlloc:
        """The allocator this ref belongs to."""
        return self._alloc

    @property
    def reg(self) -> QAlloc:
        """Alias for alloc - compatibility with Qubit.reg interface."""
        return self._alloc

    @property
    def index(self) -> int:
        """The slot index within the allocator."""
        return self._index

    @property
    def state(self) -> SlotState:
        """Current state of this slot."""
        return self._alloc.state(self._index)

    @property
    def is_prepared(self) -> bool:
        """Whether this slot is prepared and ready for gates."""
        return self.state == SlotState.PREPARED

    def __repr__(self) -> str:
        return f"QubitRef({self._alloc.name}[{self._index}])"

    def __str__(self) -> str:
        return f"{self._alloc.name}[{self._index}]"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QubitRef):
            return NotImplemented
        return self._alloc is other._alloc and self._index == other._index

    def __hash__(self) -> int:
        return hash((id(self._alloc), self._index))


class QAlloc:
    """Qubit allocator managing N qubit slots.

    Inspired by Zig's allocator pattern. Provides:
    - Hierarchical ownership (parent-child relationships)
    - Slot-based access (logical indices, not physical qubits)
    - Lifecycle tracking (unprepared/prepared states)
    - Natural scoping (unreturned allocators release to parent)

    Usage:
        # Base allocator in main
        base = QAlloc(capacity=100)

        # Child allocators partition the resource
        data = base.child(7, name="data")
        ancilla = base.child(6, name="ancilla")

        # Prepare slots before use
        data.prepare_all()

        # Access via indexing
        H(data[0])
        CNOT(data[0], data[1])

        # Measure transitions to unprepared
        result = Measure(data[0])  # data[0] now unprepared

        # Re-prepare for reuse
        data.prepare(0)
    """

    def __init__(
        self,
        capacity: int,
        *,
        name: str | None = None,
        parent: QAlloc | None = None,
        _parent_indices: list[int] | None = None,
    ) -> None:
        """Create a qubit allocator.

        Args:
            capacity: Number of qubit slots in this allocator.
            name: Optional name for this allocator (for debugging/codegen).
            parent: Parent allocator (None for base allocator).
            _parent_indices: Internal - indices reserved from parent.
        """
        if capacity < 0:
            msg = f"Capacity must be non-negative, got {capacity}"
            raise ValueError(msg)

        self._capacity = capacity
        self._name = name
        self._parent = parent
        self._parent_indices = _parent_indices or []

        # Slot states - all start unprepared
        self._slot_states: list[SlotState] = [SlotState.UNPREPARED] * capacity

        # Track which slots are reserved by children
        self._reserved: set[int] = set()

        # Track child allocators
        self._children: list[QAlloc] = []

        # Track if this allocator has been released
        self._released = False

    # --- Properties ---

    @property
    def capacity(self) -> int:
        """Total number of slots in this allocator."""
        return self._capacity

    @property
    def size(self) -> int:
        """Alias for capacity - compatibility with QReg.size interface."""
        return self._capacity

    @property
    def name(self) -> str:
        """Name of this allocator."""
        return self._name or f"alloc_{id(self)}"

    @property
    def sym(self) -> str:
        """Alias for name - compatibility with QReg.sym interface."""
        return self.name

    @property
    def parent(self) -> QAlloc | None:
        """Parent allocator, or None if this is a base allocator."""
        return self._parent

    @property
    def is_base(self) -> bool:
        """Whether this is a base (root) allocator."""
        return self._parent is None

    @property
    def available(self) -> int:
        """Number of slots not reserved by children."""
        return self._capacity - len(self._reserved)

    @property
    def is_released(self) -> bool:
        """Whether this allocator has been released."""
        return self._released

    # --- Child Allocator Creation ---

    def child(self, size: int, *, name: str | None = None) -> QAlloc:
        """Create a child allocator with `size` slots.

        Reserves `size` slots from this allocator's available pool.
        Child allocator slots start unprepared.

        Args:
            size: Number of slots for the child allocator.
            name: Optional name for the child allocator.

        Returns:
            A new child QAlloc with `size` slots.

        Raises:
            ValueError: If insufficient capacity available.
            RuntimeError: If this allocator has been released.
        """
        self._check_not_released()

        if size < 0:
            msg = f"Size must be non-negative, got {size}"
            raise ValueError(msg)

        if size > self.available:
            msg = f"Insufficient capacity: requested {size}, available {self.available}"
            raise ValueError(msg)

        # Find available indices to reserve
        available_indices = [i for i in range(self._capacity) if i not in self._reserved]
        indices_to_reserve = available_indices[:size]

        # Mark as reserved
        self._reserved.update(indices_to_reserve)

        # Create child
        child_alloc = QAlloc(
            capacity=size,
            name=name,
            parent=self,
            _parent_indices=indices_to_reserve,
        )
        self._children.append(child_alloc)

        return child_alloc

    # --- Lifecycle Operations ---

    def prepare(self, *indices: int) -> None:
        """Prepare specific slots (unprepared -> prepared).

        Requesting a qubit to be associated with a slot and initialized to |0>.

        Args:
            *indices: Slot indices to prepare.

        Raises:
            IndexError: If any index is out of range.
            ValueError: If any index is reserved by a child.
            RuntimeError: If this allocator has been released.
        """
        self._check_not_released()

        for idx in indices:
            self._check_index(idx)
            self._check_not_reserved(idx)
            self._slot_states[idx] = SlotState.PREPARED

    def prepare_all(self) -> None:
        """Prepare all slots in this allocator.

        Only prepares slots not reserved by children.

        Raises:
            RuntimeError: If this allocator has been released.
        """
        self._check_not_released()

        for i in range(self._capacity):
            if i not in self._reserved:
                self._slot_states[i] = SlotState.PREPARED

    def mark_unprepared(self, *indices: int) -> None:
        """Mark specific slots as unprepared (after measurement).

        This is called by Measure operations to transition slots.

        Args:
            *indices: Slot indices to mark unprepared.

        Raises:
            IndexError: If any index is out of range.
            RuntimeError: If this allocator has been released.
        """
        self._check_not_released()

        for idx in indices:
            self._check_index(idx)
            self._slot_states[idx] = SlotState.UNPREPARED

    def mark_all_unprepared(self) -> None:
        """Mark all slots as unprepared.

        Raises:
            RuntimeError: If this allocator has been released.
        """
        self._check_not_released()

        for i in range(self._capacity):
            if i not in self._reserved:
                self._slot_states[i] = SlotState.UNPREPARED

    # --- Slot State Queries ---

    def state(self, index: int) -> SlotState:
        """Get the state of a specific slot.

        Args:
            index: Slot index.

        Returns:
            SlotState.UNPREPARED or SlotState.PREPARED.

        Raises:
            IndexError: If index is out of range.
        """
        self._check_index(index)
        return self._slot_states[index]

    def is_prepared(self, index: int) -> bool:
        """Check if a specific slot is prepared.

        Args:
            index: Slot index.

        Returns:
            True if the slot is prepared, False otherwise.
        """
        return self.state(index) == SlotState.PREPARED

    def all_prepared(self) -> bool:
        """Check if all non-reserved slots are prepared."""
        return all(self._slot_states[i] == SlotState.PREPARED for i in range(self._capacity) if i not in self._reserved)

    def prepared_count(self) -> int:
        """Count of prepared slots."""
        return sum(
            1 for i in range(self._capacity) if i not in self._reserved and self._slot_states[i] == SlotState.PREPARED
        )

    def unprepared_count(self) -> int:
        """Count of unprepared slots."""
        return sum(
            1 for i in range(self._capacity) if i not in self._reserved and self._slot_states[i] == SlotState.UNPREPARED
        )

    # --- Slot Access ---

    def __getitem__(self, index: int) -> QubitRef:
        """Access a slot for use in gates.

        Returns a QubitRef that can be passed to gate operations.

        Args:
            index: Slot index.

        Returns:
            QubitRef for the slot.

        Raises:
            IndexError: If index is out of range.
            ValueError: If index is reserved by a child.
            RuntimeError: If this allocator has been released.
        """
        self._check_not_released()
        self._check_index(index)
        self._check_not_reserved(index)

        return QubitRef(self, index)

    def __len__(self) -> int:
        """Total capacity of this allocator."""
        return self._capacity

    def __iter__(self) -> Iterator[QubitRef]:
        """Iterate over all non-reserved slots as QubitRefs."""
        self._check_not_released()
        for i in range(self._capacity):
            if i not in self._reserved:
                yield QubitRef(self, i)

    # --- Release ---

    def release(self) -> None:
        """Explicitly release this allocator back to parent.

        Normally handled automatically by scoping, but can be called explicitly.
        All child allocators are also released.

        Raises:
            RuntimeError: If already released.
        """
        if self._released:
            msg = f"Allocator {self.name} already released"
            raise RuntimeError(msg)

        # Release all children first
        for child in self._children:
            if not child.is_released:
                child.release()

        # Clear our slots back to parent's reserved set
        if self._parent is not None:
            self._parent.unreserve_slots(self._parent_indices)
            self._parent.unregister_child(self)

        self._released = True

    # --- Parent/Child Coordination ---

    def unreserve_slots(self, indices: list[int]) -> None:
        """Release the given slot indices back to the available pool.

        Called by child allocators when they are released.
        Not intended for external use.
        """
        self._reserved -= set(indices)

    def unregister_child(self, child: QAlloc) -> None:
        """Remove a child allocator from this allocator's children list.

        Called by child allocators when they are released.
        Not intended for external use.
        """
        self._children.remove(child)

    # --- Validation Helpers ---

    def _check_not_released(self) -> None:
        """Raise if this allocator has been released."""
        if self._released:
            msg = f"Allocator {self.name} has been released"
            raise RuntimeError(msg)

    def _check_index(self, index: int) -> None:
        """Raise if index is out of range."""
        if not 0 <= index < self._capacity:
            msg = f"Slot index {index} out of range [0, {self._capacity})"
            raise IndexError(msg)

    def _check_not_reserved(self, index: int) -> None:
        """Raise if index is reserved by a child."""
        if index in self._reserved:
            msg = f"Slot {index} is reserved by a child allocator"
            raise ValueError(msg)

    # --- Representation ---

    def __repr__(self) -> str:
        parent_info = f", parent={self._parent.name}" if self._parent else ", base"
        return f"QAlloc({self.name}, capacity={self._capacity}, available={self.available}{parent_info})"

    def __str__(self) -> str:
        return self.name


# Backward compatibility: QReg as alias
# In the future, QReg could be implemented as a thin wrapper around QAlloc
# that auto-prepares all slots on creation (matching old behavior)



"""Concrete implementations of surface code patches."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.slr.qeclib.surface.layouts.rot_square_lattice import SquareRotatedLayout
from pecos.slr.qeclib.surface.patches.patch_base import BaseSurfacePatch

if TYPE_CHECKING:
    from pecos.slr.qeclib.surface.layouts.layout_base import Layout
    from pecos.slr.qeclib.surface.patches.patch_base import SurfacePatchOrientation


class RotatedSurfacePatch(BaseSurfacePatch):
    """Rotated surface patch."""

    def __init__(
        self,
        dx: int,
        dz: int,
        orientation: SurfacePatchOrientation,
        layout: Layout | None = None,
        name: str | None = None,
    ) -> None:
        """Initialize a rotated surface code patch.

        Args:
            dx: Distance of the X logical operator.
            dz: Distance of the Z logical operator.
            orientation: Patch orientation determining boundary conditions.
            layout: Optional custom layout. Uses SquareRotatedLayout by default.
            name: Optional custom name for the patch.
        """
        super().__init__(dx, dz, orientation, name)

        # TODO: Should each surface patch carry this or should it be stored somewhere for reuse...
        #       or cached somehow
        if layout is None:
            layout = SquareRotatedLayout()
        self.layout = layout
        self.stab_gens = self.layout.get_stabilizers_gens(self.dx, self.dz)

    def _calculate_qubit_count(self) -> int:
        return self.dx * self.dz


class NonRotatedSurfacePatch(BaseSurfacePatch):
    """Standard surface patch."""

    def _calculate_qubit_count(self) -> int:
        # TODO: fix for non-rotated surface code
        return self.dx * self.dz



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

r"""Rotated surface code lattice geometry.

The rotated surface code arranges data qubits on a diagonal lattice,
requiring fewer physical qubits for the same code distance compared
to the standard (non-rotated) layout.

Qubit layout for d=3 rotated code:
        0
       / \\
      1   2
     / \\ / \\
    3   4   5
     \\ / \\ /
      6   7
       \\ /
        8
"""

from dataclasses import dataclass

from pecos.qec.surface.layouts.square_lattice import StabilizerSupport


@dataclass(frozen=True)
class RotatedPosition:
    """Position in the rotated lattice coordinate system."""

    x: int
    y: int


def compute_rotated_x_stabilizers(d: int) -> list[StabilizerSupport]:
    """Compute X stabilizer supports for rotated surface code.

    Args:
        d: Code distance (must be odd >= 3)

    Returns:
        List of StabilizerSupport for X stabilizers
    """
    if d < 3 or d % 2 == 0:
        msg = f"Distance must be odd >= 3, got {d}"
        raise ValueError(msg)

    supports = []
    stab_idx = 0

    # Bulk X stabilizers (weight 4)
    for row in range(d - 1):
        for col in range(d - 1):
            if (row + col) % 2 == 0:
                q_tl = row * d + col
                q_tr = row * d + col + 1
                q_bl = (row + 1) * d + col
                q_br = (row + 1) * d + col + 1

                supports.append(
                    StabilizerSupport(
                        index=stab_idx,
                        data_qubits=(q_tl, q_tr, q_bl, q_br),
                        is_boundary=False,
                    ),
                )
                stab_idx += 1

    # Boundary X stabilizers (weight 2) - top and bottom edges
    for col in range(0, d - 1, 2):
        q1 = col
        q2 = col + 1
        supports.append(
            StabilizerSupport(
                index=stab_idx,
                data_qubits=(q1, q2),
                is_boundary=True,
            ),
        )
        stab_idx += 1

    for col in range((d - 1) % 2, d - 1, 2):
        q1 = (d - 1) * d + col
        q2 = (d - 1) * d + col + 1
        supports.append(
            StabilizerSupport(
                index=stab_idx,
                data_qubits=(q1, q2),
                is_boundary=True,
            ),
        )
        stab_idx += 1

    supports.sort(key=lambda s: s.index)
    return supports


def compute_rotated_z_stabilizers(d: int) -> list[StabilizerSupport]:
    """Compute Z stabilizer supports for rotated surface code.

    Args:
        d: Code distance (must be odd >= 3)

    Returns:
        List of StabilizerSupport for Z stabilizers
    """
    if d < 3 or d % 2 == 0:
        msg = f"Distance must be odd >= 3, got {d}"
        raise ValueError(msg)

    supports = []
    stab_idx = 0

    # Bulk Z stabilizers (weight 4)
    for row in range(d - 1):
        for col in range(d - 1):
            if (row + col) % 2 == 1:
                q_tl = row * d + col
                q_tr = row * d + col + 1
                q_bl = (row + 1) * d + col
                q_br = (row + 1) * d + col + 1

                supports.append(
                    StabilizerSupport(
                        index=stab_idx,
                        data_qubits=(q_tl, q_tr, q_bl, q_br),
                        is_boundary=False,
                    ),
                )
                stab_idx += 1

    # Boundary Z stabilizers (weight 2) - left and right edges
    for row in range(0, d - 1, 2):
        q1 = row * d
        q2 = (row + 1) * d
        supports.append(
            StabilizerSupport(
                index=stab_idx,
                data_qubits=(q1, q2),
                is_boundary=True,
            ),
        )
        stab_idx += 1

    for row in range((d - 1) % 2, d - 1, 2):
        q1 = row * d + (d - 1)
        q2 = (row + 1) * d + (d - 1)
        supports.append(
            StabilizerSupport(
                index=stab_idx,
                data_qubits=(q1, q2),
                is_boundary=True,
            ),
        )
        stab_idx += 1

    supports.sort(key=lambda s: s.index)
    return supports


def get_rotated_logical_x(d: int) -> tuple[int, ...]:
    """Get logical X operator qubits (left edge)."""
    return tuple(i * d for i in range(d))


def get_rotated_logical_z(d: int) -> tuple[int, ...]:
    """Get logical Z operator qubits (top edge)."""
    return tuple(range(d))


def rotated_id_to_position(qubit_id: int, d: int) -> tuple[int, int]:
    """Convert qubit ID to (x, y) position in rotated coordinates."""
    row = qubit_id // d
    col = qubit_id % d
    x = col * 2 + 1
    y = (d - row) * 2 - 1
    return (x, y)


def rotated_position_to_id(x: int, y: int, d: int) -> int:
    """Convert rotated position to qubit ID."""
    col = (x - 1) // 2
    row = d - (y + 1) // 2
    return row * d + col


def generate_surface_layout(
    width: int,
    height: int,
) -> tuple[list[tuple[int, int]], list[tuple[int, int]]]:
    """Generate rotated surface code layout positions.

    This is the most common surface code variant, using d^2 data qubits for
    distance d. The non-rotated variant uses more qubits; for that, use
    generate_nonrotated_surface_layout() instead.

    The rotated surface code places data qubits at odd-odd positions
    and ancilla qubits at even-even positions in the interior, with
    boundary ancillas on the edges.

    Args:
        width: Width of the patch (code distance in Z direction).
        height: Height of the patch (code distance in X direction).

    Returns:
        A tuple containing:
        - data_positions: List of (x, y) coordinates for data qubits.
        - ancilla_positions: List of (x, y) coordinates for ancilla qubits.
    """
    lattice_height = 2 * height
    lattice_width = 2 * width

    data_positions: list[tuple[int, int]] = []
    ancilla_positions: list[tuple[int, int]] = []

    for y in range(lattice_height + 1):
        for x in range(lattice_width + 1):
            if 0 < x < lattice_width and 0 < y < lattice_height:
                # Interior (no boundary stabilizers)
                if x % 2 == 1 and y % 2 == 1:
                    # Data qubit at odd-odd positions
                    data_positions.append((x, y))
                elif x % 2 == 0 and y % 2 == 0:
                    # Ancilla at even-even positions
                    ancilla_positions.append((x, y))

            elif 0 < x < lattice_width or 0 < y < lattice_height:
                # Boundary ancillas (not corners)
                if y == 0:
                    # Top edge: X stabilizers
                    if x != 0 and x % 4 == 0:
                        ancilla_positions.append((x, y))

                elif x == 0 and (y - 2) % 4 == 0:
                    # Left edge
                    ancilla_positions.append((x, y))

                if y == lattice_height:
                    # Bottom edge
                    if height % 2 == 0:
                        if x != 0 and x % 4 == 0:
                            ancilla_positions.append((x, y))
                    elif (x - 2) % 4 == 0:
                        ancilla_positions.append((x, y))

                elif x == lattice_width:
                    # Right edge
                    if width % 2 == 1:
                        if y != 0 and y % 4 == 0:
                            ancilla_positions.append((x, y))
                    elif (y - 2) % 4 == 0:
                        ancilla_positions.append((x, y))

    return data_positions, ancilla_positions



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""List type definitions for PyPHIR intermediate representation.

This module defines specialized list types for PyPHIR (Python PECOS Medium-level Intermediate Representation) including
typed lists for instructions, operations, and other quantum circuit elements.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.reps.pyphir.instr_type import Instr
from pecos.reps.pyphir.op_types import Op, QOp
from pecos.typed_list import TypedList

if TYPE_CHECKING:
    from collections.abc import Iterable


class InstrList(TypedList):
    """A list of general Instructions include Ops, Blocks, and Data."""

    _type = Instr

    def __init__(self, data: Iterable[Instr] | None = None) -> None:
        """Initialize an InstrList.

        Args:
            data: Optional iterable of Instr objects to initialize the list.
        """
        super().__init__(self._type, data)
        self.metadata = None


class OpList(InstrList):
    """A list of Operations, e.g., QOp, MOp,EMOp, etc.."""

    _type = Op

    def __init__(self, data: Iterable[Op] | None = None) -> None:
        """Initialize an OpList.

        Args:
            data: Optional iterable of Op objects to initialize the list.
        """
        super().__init__(data)


class QOpList(OpList):
    """A list of just QOps."""

    _type = QOp

    def __init__(self, data: Iterable[QOp] | None = None) -> None:
        """Initialize a QOpList.

        Args:
            data: Optional iterable of QOp (quantum operation) objects to initialize the list.
        """
        super().__init__(data)



"""HUGR to LLVM compiler integration for PECOS.

This module provides a Python interface to compile HUGR files to LLVM IR using
the working quantum compilation pipeline from quantum-compilation-examples.
"""

import contextlib
import os
import shutil
import subprocess
import tempfile
from pathlib import Path


class HugrLlvmCompiler:
    """Compiler that converts HUGR files to LLVM IR with quantum operations.

    This uses the working quantum compilation pipeline from the
    quantum-compilation-examples project.
    """

    def __init__(self, hugr_llvm_binary: Path | None = None) -> None:
        """Initialize the HUGR→LLVM compiler.

        Args:
            hugr_llvm_binary: Path to the hugr-to-llvm compiler binary.
                             If None, will try to find it automatically.
        """
        self.hugr_llvm_binary = hugr_llvm_binary or self._find_hugr_llvm_binary()
        self._temp_dir = None

    def _find_hugr_llvm_binary(self) -> Path | None:
        """Find the hugr-to-llvm compiler binary."""
        # Check common locations relative to PECOS
        base_dir = Path(__file__).parent.parent.parent.parent.parent.parent

        possible_paths = [
            # In quantum-compilation-examples
            base_dir / "quantum-compilation-examples/hugr_quantum_llvm/target/release/hugr_quantum_llvm",
            base_dir / "quantum-compilation-examples/hugr_quantum_llvm/target/debug/hugr_quantum_llvm",
            # Built versions
            Path("./hugr_quantum_llvm"),
            Path("hugr_quantum_llvm"),
        ]

        for path in possible_paths:
            if path.exists() and os.access(path, os.X_OK):
                return path.resolve()

        return None

    def compile_hugr_to_qis(
        self,
        hugr_bytes: bytes,
        output_file: Path | None = None,
    ) -> str:
        """Compile HUGR bytes to LLVM IR.

        Args:
            hugr_bytes: HUGR package as bytes
            output_file: Optional output file path

        Returns:
            LLVM IR as string (HUGR convention)

        Raises:
            RuntimeError: If compilation fails
        """
        if not self.hugr_llvm_binary:
            msg = "HUGR→LLVM compiler not found. Build it from quantum-compilation-examples/hugr_quantum_llvm/"
            raise RuntimeError(
                msg,
            )

        # Create temporary directory
        if self._temp_dir is None:
            self._temp_dir = tempfile.mkdtemp(prefix="hugr_llvm_")

        temp_path = Path(self._temp_dir)

        # Write HUGR to temporary file
        hugr_file = temp_path / "input.hugr"
        with Path(hugr_file).open("wb") as f:
            f.write(hugr_bytes)

        # Determine output file
        llvm_file = temp_path / "output.ll" if output_file is None else output_file

        # Run the compiler
        try:
            cmd = [
                str(self.hugr_llvm_binary),
                str(hugr_file),
                str(llvm_file),
                "hugr",  # Always use HUGR convention
            ]

            subprocess.run(
                cmd,
                check=True,
                capture_output=True,
                text=True,
            )

            # Read the generated LLVM IR
            with Path(llvm_file).open() as f:
                return f.read()

        except subprocess.CalledProcessError as e:
            msg = f"HUGR→LLVM compilation failed: {e.stderr}"
            raise RuntimeError(msg) from e
        except FileNotFoundError as e:
            msg = f"Compiler binary not found: {self.hugr_llvm_binary}"
            raise RuntimeError(msg) from e

    def is_available(self) -> bool:
        """Check if the HUGR→LLVM compiler is available."""
        return self.hugr_llvm_binary is not None and self.hugr_llvm_binary.exists()

    def cleanup(self) -> None:
        """Clean up temporary files."""
        if hasattr(self, "_temp_dir") and self._temp_dir and Path(self._temp_dir).exists():

            shutil.rmtree(self._temp_dir)
            self._temp_dir = None

    def __del__(self) -> None:
        """Cleanup on destruction."""
        with contextlib.suppress(Exception):
            self.cleanup()


def compile_hugr_bytes_to_llvm(
    hugr_bytes: bytes,
) -> str:
    """Convenience function to compile HUGR bytes to LLVM IR.

    Args:
        hugr_bytes: HUGR package as bytes

    Returns:
        LLVM IR as string (HUGR convention)
    """
    compiler = HugrLlvmCompiler()
    try:
        return compiler.compile_hugr_to_qis(hugr_bytes)
    finally:
        compiler.cleanup()


def build_hugr_llvm_compiler() -> bool:
    """Build the HUGR→LLVM compiler if source is available.

    Returns:
        True if build succeeded, False otherwise
    """
    # Find the source directory
    base_dir = Path(__file__).parent.parent.parent.parent.parent.parent
    source_dir = base_dir / "quantum-compilation-examples/hugr_quantum_llvm"

    if not source_dir.exists():
        print(f"Source directory not found: {source_dir}")
        return False

    try:
        # Build in release mode
        subprocess.run(
            ["cargo", "build", "--release"],
            cwd=source_dir,
            check=True,
            capture_output=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        print(f"[ERROR] Build failed: {e.stderr}")
        return False
    except FileNotFoundError:
        print("[ERROR] cargo not found - install Rust toolchain")
        return False
    else:
        binary_path = source_dir / "target/release/hugr_quantum_llvm"
        if binary_path.exists():
            print(f"[PASS] Built HUGR->LLVM compiler: {binary_path}")
            return True
        print("[ERROR] Build succeeded but binary not found")
        return False


if __name__ == "__main__":
    # Test the compiler
    print("Testing HUGR->LLVM compiler...")

    compiler = HugrLlvmCompiler()

    if compiler.is_available():
        print(f"[PASS] Compiler available: {compiler.hugr_llvm_binary}")
    else:
        print("[ERROR] Compiler not available")
        print("Attempting to build...")
        if build_hugr_llvm_compiler():
            compiler = HugrLlvmCompiler()  # Re-initialize to find new binary
            if compiler.is_available():
                print(f"[PASS] Compiler now available: {compiler.hugr_llvm_binary}")
        else:
            print("[ERROR] Failed to build compiler")



"""Transversal two-qubit logical gates for the Steane 7-qubit code.

This module provides transversal two-qubit logical gate implementations for the Steane 7-qubit code, enabling
fault-tolerant operations between logical qubits while preserving error correction properties.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from pecos.slr import Barrier, Block, Comment, QReg
from pecos.slr.qeclib import qubit


class CX(Block):
    """Transversal logical CX gate for Steane code.

    This class implements a transversal logical controlled-X gate between
    two logical qubits encoded in the Steane code.
    """

    def __init__(self, q1: QReg, q2: QReg, *, barrier: bool = True) -> None:
        """Initialize a transversal logical CX gate on two Steane code logical qubits.

        Args:
            q1: First quantum register containing exactly 7 qubits (control).
            q2: Second quantum register containing exactly 7 qubits (target).
            barrier: Whether to include barriers before and after the gate operation.
                Defaults to True.

        Raises:
            Exception: If either quantum register does not contain exactly 7 qubits.
        """
        if len(q1.elems) != 7:
            msg = f"Size of register {len(q1.elems)} != 7"
            raise Exception(msg)

        if len(q2.elems) != 7:
            msg = f"Size of register {len(q2.elems)} != 7"
            raise Exception(msg)

        super().__init__()
        self.extend(
            Comment("Transversal Logical CX"),
        )
        if barrier:
            self.extend(
                Barrier(q1, q2),
            )
        self.extend(
            qubit.CX(
                (q1[0], q2[0]),
                (q1[1], q2[1]),
                (q1[2], q2[2]),
                (q1[3], q2[3]),
                (q1[4], q2[4]),
                (q1[5], q2[5]),
                (q1[6], q2[6]),
            ),
        )
        if barrier:
            self.extend(
                Barrier(q1, q2),
            )


class CY(Block):
    """Transversal logical CY gate for Steane code.

    This class implements a transversal logical controlled-Y gate between
    two logical qubits encoded in the Steane code.
    """

    def __init__(self, q1: QReg, q2: QReg) -> None:
        """Initialize a transversal logical CY gate on two Steane code logical qubits.

        Args:
            q1: First quantum register containing exactly 7 qubits (control).
            q2: Second quantum register containing exactly 7 qubits (target).

        Raises:
            Exception: If either quantum register does not contain exactly 7 qubits.
        """
        if len(q1.elems) != 7:
            msg = f"Size of register {len(q1.elems)} != 7"
            raise Exception(msg)

        if len(q2.elems) != 7:
            msg = f"Size of register {len(q2.elems)} != 7"
            raise Exception(msg)

        super().__init__(
            Comment("Transversal Logical CY"),
            Barrier(q1, q2),
            qubit.CY(
                (q1[0], q2[0]),
                (q1[1], q2[1]),
                (q1[2], q2[2]),
                (q1[3], q2[3]),
                (q1[4], q2[4]),
                (q1[5], q2[5]),
                (q1[6], q2[6]),
            ),
            Barrier(q1, q2),
        )


class CZ(Block):
    """Transversal logical CZ gate for Steane code.

    This class implements a transversal logical controlled-Z gate between
    two logical qubits encoded in the Steane code.
    """

    def __init__(self, q1: QReg, q2: QReg) -> None:
        """Initialize a transversal logical CZ gate on two Steane code logical qubits.

        Args:
            q1: First quantum register containing exactly 7 qubits.
            q2: Second quantum register containing exactly 7 qubits.

        Raises:
            Exception: If either quantum register does not contain exactly 7 qubits.
        """
        if len(q1.elems) != 7:
            msg = f"Size of register {len(q1.elems)} != 7"
            raise Exception(msg)

        if len(q2.elems) != 7:
            msg = f"Size of register {len(q2.elems)} != 7"
            raise Exception(msg)

        super().__init__(
            Comment("Transversal Logical CZ"),
            Barrier(q1, q2),
            qubit.CZ(
                (q1[0], q2[0]),
                (q1[1], q2[1]),
                (q1[2], q2[2]),
                (q1[3], q2[3]),
                (q1[4], q2[4]),
                (q1[5], q2[5]),
                (q1[6], q2[6]),
            ),
            Barrier(q1, q2),
        )


class SZZ(Block):
    """Transversal logical SZZ gate for Steane code.

    This class implements a transversal logical SZZ interaction gate between
    two logical qubits encoded in the Steane code.
    """

    def __init__(self, q1: QReg, q2: QReg) -> None:
        """Initialize a transversal logical SZZ gate on two Steane code logical qubits.

        Args:
            q1: First quantum register containing exactly 7 qubits.
            q2: Second quantum register containing exactly 7 qubits.

        Raises:
            Exception: If either quantum register does not contain exactly 7 qubits.
        """
        if len(q1.elems) != 7:
            msg = f"Size of register {len(q1.elems)} != 7"
            raise Exception(msg)

        if len(q2.elems) != 7:
            msg = f"Size of register {len(q2.elems)} != 7"
            raise Exception(msg)

        super().__init__(
            Comment("Transversal Logical SZZ"),
            Barrier(q1, q2),
            qubit.SZZ(
                (q1[0], q2[0]),
                (q1[1], q2[1]),
                (q1[2], q2[2]),
                (q1[3], q2[3]),
                (q1[4], q2[4]),
                (q1[5], q2[5]),
                (q1[6], q2[6]),
            ),
            Barrier(q1, q2),
        )


class SZZdg(Block):
    """Transversal logical SZZdg gate for Steane code.

    This class implements a transversal logical SZZ interaction gate between
    two logical qubits encoded in the Steane code.
    """

    def __init__(self, q1: QReg, q2: QReg) -> None:
        """Initialize a transversal logical SZZdg gate on two Steane code logical qubits.

        Args:
            q1: First quantum register containing exactly 7 qubits.
            q2: Second quantum register containing exactly 7 qubits.

        Raises:
            Exception: If either quantum register does not contain exactly 7 qubits.
        """
        if len(q1.elems) != 7:
            msg = f"Size of register {len(q1.elems)} != 7"
            raise Exception(msg)

        if len(q2.elems) != 7:
            msg = f"Size of register {len(q2.elems)} != 7"
            raise Exception(msg)

        super().__init__(
            Comment("Transversal Logical SZZdg"),
            Barrier(q1, q2),
            qubit.SZZdg(
                (q1[0], q2[0]),
                (q1[1], q2[1]),
                (q1[2], q2[2]),
                (q1[3], q2[3]),
                (q1[4], q2[4]),
                (q1[5], q2[5]),
                (q1[6], q2[6]),
            ),
            Barrier(q1, q2),
        )



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""Internal compilation utilities for PECOS.

This module contains internal utilities for compiling quantum programs from
various formats (Guppy, HUGR) to executable formats (LLVM/QIS).

These are implementation details and should not be imported directly by users.
"""

from pecos._compilation.guppy import GuppyFrontend, compile_guppy_to_qir, guppy_to_hugr
from pecos._compilation.hugr_llvm import HugrLlvmCompiler, compile_hugr_bytes_to_llvm

__all__ = [
    "GuppyFrontend",
    "HugrLlvmCompiler",
    "compile_guppy_to_qir",
    "compile_hugr_bytes_to_llvm",
    "guppy_to_hugr",
]
